# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import division

import numpy as np
import odl
import pytest
from odl.core.util import is_real_dtype
from odl.core.util.numerics import (
    _SUPPORTED_RESIZE_PAD_MODES, apply_on_boundary, binning,
    fast_1d_tensor_mult, resize_array)
from odl.core.util.testutils import (
    all_almost_equal, all_equal, dtype_tol, simple_fixture)

# --- pytest fixtures --- #


paddings = list(_SUPPORTED_RESIZE_PAD_MODES)
paddings.remove('constant')
paddings.extend([('constant', 0), ('constant', 1)])
padding_ids = [" pad_mode='{}'-{} ".format(*p)
               if isinstance(p, tuple)
               else " pad_mode='{}' ".format(p)
               for p in paddings]


@pytest.fixture(scope="module", ids=padding_ids, params=paddings)
def padding(request):
    if isinstance(request.param, tuple):
        pad_mode, pad_const = request.param
    else:
        pad_mode = request.param
        pad_const = 0

    return pad_mode, pad_const


variant = simple_fixture('variant', ['extend', 'restrict', 'mixed'])


@pytest.fixture(scope="module")
def resize_setup(padding, variant):
    array_in = [[1, 2, 3, 4],
                [5, 6, 7, 8],
                [9, 10, 11, 12]]

    pad_mode, pad_const = padding

    if variant == 'extend':
        newshp = (4, 7)
        offset = (1, 2)

        if pad_mode == 'constant' and pad_const == 0:
            true_out = [[0, 0, 0, 0, 0, 0, 0],
                        [0, 0, 1, 2, 3, 4, 0],
                        [0, 0, 5, 6, 7, 8, 0],
                        [0, 0, 9, 10, 11, 12, 0]]
        elif pad_mode == 'constant' and pad_const == 1:
            true_out = [[1, 1, 1, 1, 1, 1, 1],
                        [1, 1, 1, 2, 3, 4, 1],
                        [1, 1, 5, 6, 7, 8, 1],
                        [1, 1, 9, 10, 11, 12, 1]]
        elif pad_mode == 'periodic':
            true_out = [[11, 12, 9, 10, 11, 12, 9],
                        [3, 4, 1, 2, 3, 4, 1],
                        [7, 8, 5, 6, 7, 8, 5],
                        [11, 12, 9, 10, 11, 12, 9]]
        elif pad_mode == 'symmetric':
            true_out = [[7, 6, 5, 6, 7, 8, 7],
                        [3, 2, 1, 2, 3, 4, 3],
                        [7, 6, 5, 6, 7, 8, 7],
                        [11, 10, 9, 10, 11, 12, 11]]
        elif pad_mode == 'order0':
            true_out = [[1, 1, 1, 2, 3, 4, 4],
                        [1, 1, 1, 2, 3, 4, 4],
                        [5, 5, 5, 6, 7, 8, 8],
                        [9, 9, 9, 10, 11, 12, 12]]
        elif pad_mode == 'order1':
            true_out = [[-5, -4, -3, -2, -1, 0, 1],
                        [-1, 0, 1, 2, 3, 4, 5],
                        [3, 4, 5, 6, 7, 8, 9],
                        [7, 8, 9, 10, 11, 12, 13]]

    elif variant == 'restrict':
        newshp = (2, 1)
        offset = (1, 2)

        true_out = [[7],
                    [11]]

    elif variant == 'mixed':
        newshp = (2, 7)
        offset = (1, 2)

        if pad_mode == 'constant' and pad_const == 0:
            true_out = [[0, 0, 5, 6, 7, 8, 0],
                        [0, 0, 9, 10, 11, 12, 0]]
        elif pad_mode == 'constant' and pad_const == 1:
            true_out = [[1, 1, 5, 6, 7, 8, 1],
                        [1, 1, 9, 10, 11, 12, 1]]
        elif pad_mode == 'periodic':
            true_out = [[7, 8, 5, 6, 7, 8, 5],
                        [11, 12, 9, 10, 11, 12, 9]]
        elif pad_mode == 'symmetric':
            true_out = [[7, 6, 5, 6, 7, 8, 7],
                        [11, 10, 9, 10, 11, 12, 11]]
        elif pad_mode == 'order0':
            true_out = [[5, 5, 5, 6, 7, 8, 8],
                        [9, 9, 9, 10, 11, 12, 12]]
        elif pad_mode == 'order1':
            true_out = [[3, 4, 5, 6, 7, 8, 9],
                        [7, 8, 9, 10, 11, 12, 13]]
        else:
            raise ValueError('unknown param')

    else:
        raise ValueError('unknown variant')

    return pad_mode, pad_const, newshp, offset, array_in, true_out


bin_size = simple_fixture(
    'bin_size', [0, 1, 2, 3, (1, 2), (4, 2), (8, 1), (2, 0), (2, 2, 2)]
)
bin_reduction = simple_fixture(
    'bin_reduction', [np.sum, np.max, np.mean], fmt=' {name}={value.__name__} '
)
bin_dtype = simple_fixture("bin_dtype", ['float64', 'int64'])


@pytest.fixture(scope="module")
def binning_setup(bin_size, bin_reduction, bin_dtype):
    arr = np.array([
        [0, 1, -1, 2, 0, 6],
        [-4, 2, 5, 2, -2, -1],
        [0, 0, 1, 0, 1, -1],
        [1, -1, 1, -1, 1, -1],
    ], dtype=bin_dtype)

    if bin_reduction == np.sum:
        out_dtype = bin_dtype

        if bin_size == 1:
            true_out_arr = arr.astype(out_dtype)
        elif bin_size == 2:
            true_out_arr = np.array([
                [-1, 8, 3],
                [0, 1, 0],
            ], dtype=out_dtype)
        elif bin_size == (1, 2):
            true_out_arr = np.array([
                [1, 1, 6],
                [-2, 7, -3],
                [0, 1, 0],
                [0, 0, 0],
            ], dtype=out_dtype)
        elif bin_size == (4, 2):
            true_out_arr = np.array([[-1, 9, 3]], dtype=out_dtype)
        else:
            true_out_arr = None

    elif bin_reduction == np.max:
        out_dtype = bin_dtype

        if bin_size == 1:
            true_out_arr = arr.astype(out_dtype)
        elif bin_size == 2:
            true_out_arr = np.array([
                [2, 5, 6],
                [1, 1, 1],
            ], dtype=out_dtype)
        elif bin_size == (1, 2):
            true_out_arr = np.array([
                [1, 2, 6],
                [2, 5, -1],
                [0, 1, 1],
                [1, 1, 1],
            ], dtype=out_dtype)
        elif bin_size == (4, 2):
            true_out_arr = np.array([[2, 5, 6]], dtype=out_dtype)
        else:
            true_out_arr = None

    elif bin_reduction == np.mean:
        out_dtype = np.result_type(bin_dtype, 1.0)

        if bin_size == 1:
            true_out_arr = arr.astype(out_dtype)
        elif bin_size == 2:
            true_out_arr = np.array([
                [-0.25, 2, 0.75],
                [0, 0.25, 0],
            ], dtype=out_dtype)
        elif bin_size == (1, 2):
            true_out_arr = np.array([
                [0.5, 0.5, 3],
                [-1, 3.5, -1.5],
                [0, 0.5, 0],
                [0, 0, 0],
            ], dtype=out_dtype)
        elif bin_size == (4, 2):
            true_out_arr = np.array([[-0.125, 1.125, 0.375]], dtype=out_dtype)
        else:
            true_out_arr = None

    else:
        assert False, "unknown reduction"

    return bin_size, bin_reduction, bin_dtype, arr, true_out_arr


# --- apply_on_boundary --- #


def test_apply_on_boundary_default():

    # 1d
    arr = np.ones(5)
    result = apply_on_boundary(arr, lambda x: x * 2)
    assert all_equal(arr, [1, 1, 1, 1, 1])
    assert all_equal(result, [2, 1, 1, 1, 2])

    # 3d
    arr = np.ones((3, 4, 5))
    result = apply_on_boundary(arr, lambda x: x * 2)
    true_arr = 2 * np.ones((3, 4, 5))
    true_arr[1:-1, 1:-1, 1:-1] = 1
    assert all_equal(result, true_arr)


def test_apply_on_boundary_func_sequence_2d():

    arr = np.ones((3, 5))
    result = apply_on_boundary(arr, [lambda x: x * 2, lambda x: x * 3])
    assert all_equal(result, [[2, 2, 2, 2, 2],
                              [3, 1, 1, 1, 3],
                              [2, 2, 2, 2, 2]])


def test_apply_on_boundary_multiple_times_2d():

    arr = np.ones((3, 5))
    result = apply_on_boundary(arr, lambda x: x * 2, only_once=False)
    assert all_equal(result, [[4, 2, 2, 2, 4],
                              [2, 1, 1, 1, 2],
                              [4, 2, 2, 2, 4]])


def test_apply_on_boundary_which_boundaries():

    # 1d
    arr = np.ones(5)
    which = ((True, False),)
    result = apply_on_boundary(arr, lambda x: x * 2, which_boundaries=which)
    assert all_equal(result, [2, 1, 1, 1, 1])

    # 2d
    arr = np.ones((3, 5))
    which = ((True, False), True)
    result = apply_on_boundary(arr, lambda x: x * 2, which_boundaries=which)
    assert all_equal(result, [[2, 2, 2, 2, 2],
                              [2, 1, 1, 1, 2],
                              [2, 1, 1, 1, 2]])


def test_apply_on_boundary_which_boundaries_multiple_times_2d():

    # 2d
    arr = np.ones((3, 5))
    which = ((True, False), True)
    result = apply_on_boundary(arr, lambda x: x * 2, which_boundaries=which,
                               only_once=False)
    assert all_equal(result, [[4, 2, 2, 2, 4],
                              [2, 1, 1, 1, 2],
                              [2, 1, 1, 1, 2]])


def test_apply_on_boundary_axis_order_2d():

    arr = np.ones((3, 5))
    axis_order = (-1, -2)
    result = apply_on_boundary(arr, [lambda x: x * 3, lambda x: x * 2],
                               axis_order=axis_order)
    assert all_equal(result, [[3, 2, 2, 2, 3],
                              [3, 1, 1, 1, 3],
                              [3, 2, 2, 2, 3]])


# --- fast_1d_tensor_mult --- #


def test_fast_1d_tensor_mult():

    # Full multiplication
    def simple_mult_3(x, y, z):
        return x[:, None, None] * y[None, :, None] * z[None, None, :]

    shape = (2, 3, 4)
    x, y, z = (np.arange(size, dtype='float64') for size in shape)
    true_result = simple_mult_3(x, y, z)

    # Standard call
    test_arr = np.ones(shape)
    out = fast_1d_tensor_mult(test_arr, [x, y, z])
    assert all_equal(out, true_result)
    assert all_equal(test_arr, np.ones(shape))  # no changes to input

    test_arr = np.ones(shape)
    out = fast_1d_tensor_mult(test_arr, [x, y, z], axes=(0, 1, 2))
    assert all_equal(out, true_result)

    # In-place with both same and different array as input
    test_arr = np.ones(shape)
    fast_1d_tensor_mult(test_arr, [x, y, z], out=test_arr)
    assert all_equal(test_arr, true_result)

    test_arr = np.ones(shape)
    out = np.empty(shape)
    fast_1d_tensor_mult(test_arr, [x, y, z], out=out)
    assert all_equal(out, true_result)

    # Different orderings
    test_arr = np.ones(shape)
    out = fast_1d_tensor_mult(test_arr, [y, x, z], axes=(1, 0, 2))
    assert all_equal(out, true_result)

    test_arr = np.ones(shape)
    out = fast_1d_tensor_mult(test_arr, [x, z, y], axes=(0, 2, 1))
    assert all_equal(out, true_result)

    test_arr = np.ones(shape)
    out = fast_1d_tensor_mult(test_arr, [z, x, y], axes=(2, 0, 1))
    assert all_equal(out, true_result)

    # More arrays than dimensions also ok with explicit axes
    test_arr = np.ones(shape)
    out = fast_1d_tensor_mult(test_arr, [z, x, y, np.ones(3)],
                              axes=(2, 0, 1, 1))
    assert all_equal(out, true_result)

    # Squeezable or extendable arrays also possible
    test_arr = np.ones(shape)
    out = fast_1d_tensor_mult(test_arr, [x, y, z[None, :]])
    assert all_equal(out, true_result)

    shape = (1, 3, 4)
    test_arr = np.ones(shape)
    out = fast_1d_tensor_mult(test_arr, [2, y, z])
    assert all_equal(out, simple_mult_3(np.ones(1) * 2, y, z))

    # Reduced multiplication, axis 0 not contained
    def simple_mult_2(y, z, nx):
        return np.ones((nx, 1, 1)) * y[None, :, None] * z[None, None, :]

    shape = (2, 3, 4)
    x, y, z = (np.arange(size, dtype='float64') for size in shape)
    true_result = simple_mult_2(y, z, 2)

    test_arr = np.ones(shape)
    out = fast_1d_tensor_mult(test_arr, [y, z], axes=(1, 2))
    assert all_equal(out, true_result)

    test_arr = np.ones(shape)
    out = fast_1d_tensor_mult(test_arr, [z, y], axes=(2, 1))
    assert all_equal(out, true_result)


def test_fast_1d_tensor_mult_error():

    shape = (2, 3, 4)
    test_arr = np.ones(shape)
    x, y, z = (np.arange(size, dtype='float64') for size in shape)

    # No ndarray to operate on
    with pytest.raises(ValueError):
        fast_1d_tensor_mult([[0, 0], [0, 0]], [x, x])

    # No 1d arrays given
    with pytest.raises(ValueError):
        fast_1d_tensor_mult(test_arr, [])

    # Length or dimension mismatch
    with pytest.raises(ValueError):
        fast_1d_tensor_mult(test_arr, [x, y])

    with pytest.raises(ValueError):
        fast_1d_tensor_mult(test_arr, [x, y], (1, 2, 0))

    with pytest.raises(ValueError):
        fast_1d_tensor_mult(test_arr, [x, x, y, z])

    # Axes out of bounds
    with pytest.raises(ValueError):
        fast_1d_tensor_mult(test_arr, [x, y, z], (1, 2, 3))

    with pytest.raises(ValueError):
        fast_1d_tensor_mult(test_arr, [x, y, z], (-2, -3, -4))

    # Other than 1d arrays
    with pytest.raises(ValueError):
        fast_1d_tensor_mult(test_arr, [x, y, np.ones((4, 2))], (-2, -3, -4))


# --- resize_array --- #


def test_resize_array_fwd(resize_setup, odl_floating_dtype):
    dtype = odl_floating_dtype
    pad_mode, pad_const, newshp, offset, array_in, true_out = resize_setup
    array_in = np.array(array_in, dtype=dtype)
    true_out = np.array(true_out, dtype=dtype)

    resized = resize_array(array_in, newshp, offset, pad_mode, pad_const,
                           direction='forward')
    out = np.empty(newshp, dtype=dtype)
    resize_array(array_in, newshp, offset, pad_mode, pad_const,
                 direction='forward', out=out)

    assert np.array_equal(resized, true_out)
    assert np.array_equal(out, true_out)


def test_resize_array_adj(resize_setup, odl_floating_dtype):
    dtype = odl_floating_dtype
    pad_mode, pad_const, newshp, offset, array, _ = resize_setup

    if pad_const != 0:
        # Not well-defined
        return

    array = np.array(array, dtype=dtype)
    if is_real_dtype(dtype):
        other_arr = np.random.uniform(-10, 10, size=newshp)
    else:
        other_arr = (np.random.uniform(-10, 10, size=newshp) +
                     1j * np.random.uniform(-10, 10, size=newshp))

    resized = resize_array(array, newshp, offset, pad_mode, pad_const,
                           direction='forward')
    resized_adj = resize_array(other_arr, array.shape, offset, pad_mode,
                               pad_const, direction='adjoint')

    assert (np.vdot(resized.ravel(), other_arr.ravel()) ==
            pytest.approx(np.vdot(array.ravel(), resized_adj.ravel()),
                          rel=dtype_tol(dtype)))


def test_resize_array_corner_cases(odl_scalar_dtype, padding):
    # Test extreme cases of resizing that are still valid for several
    # `pad_mode`s
    dtype = odl_scalar_dtype
    pad_mode, pad_const = padding

    # Test array
    arr = np.arange(12, dtype=dtype).reshape((3, 4))

    # Resize to and from 0 total size
    squashed_arr = resize_array(arr, (3, 0), pad_mode=pad_mode)
    assert squashed_arr.size == 0

    squashed_arr = resize_array(arr, (0, 0), pad_mode=pad_mode)
    assert squashed_arr.size == 0

    if pad_mode == 'constant':
        # Blowing up from size 0 only works with constant padding
        true_blownup_arr = np.empty_like(arr)
        true_blownup_arr.fill(pad_const)

        blownup_arr = resize_array(np.ones((3, 0), dtype=dtype), (3, 4),
                                   pad_mode=pad_mode, pad_const=pad_const)
        assert np.array_equal(blownup_arr, true_blownup_arr)

        blownup_arr = resize_array(np.ones((0, 0), dtype=dtype), (3, 4),
                                   pad_mode=pad_mode, pad_const=pad_const)
        assert np.array_equal(blownup_arr, true_blownup_arr)

    # Resize from 0 axes to 0 axes
    zero_axes_arr = resize_array(np.array(0, dtype=dtype), (),
                                 pad_mode=pad_mode)
    assert zero_axes_arr == np.array(0, dtype=dtype)

    if pad_mode == 'periodic':
        # Resize with periodic padding, using all values from the original
        # array on both sides
        max_per_shape = (9, 12)
        res_arr = resize_array(arr, max_per_shape, pad_mode='periodic',
                               offset=arr.shape)
        assert np.array_equal(res_arr, np.tile(arr, (3, 3)))

    elif pad_mode == 'symmetric':
        # Symmetric padding, maximum number is one less compared to periodic
        # padding since the boundary value is not repeated
        arr = np.arange(6).reshape((2, 3))
        max_sym_shape = (4, 7)
        res_arr = resize_array(arr, max_sym_shape, pad_mode='symmetric',
                               offset=[1, 2])
        true_arr = np.array(
            [[5, 4, 3, 4, 5, 4, 3],
             [2, 1, 0, 1, 2, 1, 0],
             [5, 4, 3, 4, 5, 4, 3],
             [2, 1, 0, 1, 2, 1, 0]])
        assert np.array_equal(res_arr, true_arr)


def test_resize_array_raise():

    arr_1d = np.arange(6)
    arr_2d = np.arange(6).reshape((2, 3))

    # Shape not a sequence
    with pytest.raises(TypeError):
        resize_array(arr_1d, 19)

    # out given, but not an ndarray
    with pytest.raises(AttributeError):
        resize_array(arr_1d, (10,), out=[])

    # out has wrong shape
    with pytest.raises(ValueError):
        out = np.empty((4, 5))
        resize_array(arr_2d, (5, 5), out=out)

    # Input and output arrays differ in dimensionality
    with pytest.raises(ValueError):
        out = np.empty((4, 5))
        resize_array(arr_1d, (4, 5))
    with pytest.raises(ValueError):
        out = np.empty((4, 5))
        resize_array(arr_1d, (4, 5), out=out)

    # invalid pad mode
    with pytest.raises(ValueError):
        resize_array(arr_1d, (10,), pad_mode='madeup_mode')

    # padding constant cannot be cast to output data type
    with pytest.raises(ValueError):
        resize_array(arr_1d, (10,), pad_const=1.5)  # arr_1d has dtype int
    with pytest.raises(TypeError):
        arr_1d_float = arr_1d.astype(float)
        resize_array(arr_1d_float, (10,), pad_const=1.0j)

    # Too few entries for order 0 or 1 padding modes
    empty_arr = np.ones((3, 0))
    with pytest.raises(ValueError):
        resize_array(empty_arr, (3, 1), pad_mode='order0')

    small_arr = np.ones((3, 1))
    with pytest.raises(ValueError):
        resize_array(small_arr, (3, 3), pad_mode='order1')

    # Too large padding sizes for symmetric
    small_arr = np.ones((3, 1))
    with pytest.raises(ValueError):
        resize_array(small_arr, (3, 2), pad_mode='symmetric')
    with pytest.raises(ValueError):
        resize_array(small_arr, (6, 1), pad_mode='symmetric')
    with pytest.raises(ValueError):
        resize_array(small_arr, (4, 3), offset=(0, 1), pad_mode='symmetric')

    # Too large padding sizes for periodic
    small_arr = np.ones((3, 1))
    with pytest.raises(ValueError):
        resize_array(small_arr, (3, 3), pad_mode='periodic')
    with pytest.raises(ValueError):
        resize_array(small_arr, (7, 1), pad_mode='periodic')
    with pytest.raises(ValueError):
        resize_array(small_arr, (3, 4), offset=(0, 1), pad_mode='periodic')


def test_binning(binning_setup):
    """Validate the ``binning`` function."""
    bin_size, bin_reduction, _, arr, true_out_arr = binning_setup

    # Error scenarios
    bin_sizes = bin_size if isinstance(bin_size, tuple) else [bin_size] * 2
    if 0 in bin_sizes:
        # 0 not allowed
        with pytest.raises(ValueError):
            binning(arr, bin_size, bin_reduction)
        return
    if len(bin_sizes) != arr.ndim:
        with pytest.raises(ValueError):
            binning(arr, bin_size, bin_reduction)
        return
    if any(b > n for b, n in zip(bin_sizes, arr.shape)):
        # Bin size larger than array size not allowed
        with pytest.raises(ValueError):
            binning(arr, bin_size, bin_reduction)
        return
    if any(n % b != 0 for b, n in zip(bin_sizes, arr.shape)):
        # Division remainder of size/bin_size not allowed
        with pytest.raises(ValueError):
            binning(arr, bin_size, bin_reduction)
        return

    # Success scenarios
    out_arr = binning(arr, bin_size, bin_reduction)
    assert all_almost_equal(out_arr, true_out_arr)
    assert out_arr.dtype == true_out_arr.dtype


def test_binning_corner_cases():
    # 0-dim
    arr = np.array(1.0)
    assert binning(arr, 2) == 1.0  # ok since there are no axes
    with pytest.raises(ValueError):
        binning(arr, [2])  # not ok, too many binning values

    # 0 in shape, never ok
    arr = np.ones((4, 0))
    with pytest.raises(ValueError):
        binning(arr, 2)
    with pytest.raises(ValueError):
        binning(arr, [1, 1])


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Gradient-based optimization schemes."""

from __future__ import print_function, division, absolute_import
import math 

from odl.solvers.util import ConstantLineSearch

from odl.core.array_API_support import sqrt

__all__ = ('steepest_descent', 'adam')


# TODO: update all docs


def steepest_descent(f, x, line_search=1.0, maxiter=1000, tol=1e-16,
                     projection=None, callback=None):
    r"""Steepest descent method to minimize an objective function.

    General implementation of steepest decent (also known as gradient
    decent) for solving

    .. math::
        \min f(x)

    The algorithm is intended for unconstrained problems. It needs line
    search in order guarantee convergence. With appropriate line search,
    it can also be used for constrained problems where one wants to
    minimize over some given set :math:`C`. This can be done by defining
    :math:`f(x) = \infty` for :math:`x\\not\\in C`, or by providing a
    ``projection`` function that projects the iterates on :math:`C`.

    The algorithm is described in [BV2004], section 9.3--9.4
    (`book available online
    <http://stanford.edu/~boyd/cvxbook/bv_cvxbook.pdf>`_),
    [GNS2009], Section 12.2, and wikipedia
    `Gradient_descent
    <https://en.wikipedia.org/wiki/Gradient_descent>`_.

    Parameters
    ----------
    f : `Functional`
        Goal functional. Needs to have ``f.gradient``.
    x : ``f.domain`` element
        Starting point of the iteration
    line_search : float or `LineSearch`, optional
        Strategy to choose the step length. If a float is given, uses it as a
        fixed step length.
    maxiter : int, optional
        Maximum number of iterations.
    tol : float, optional
        Tolerance that should be used for terminating the iteration.
    projection : callable, optional
        Function that can be used to modify the iterates in each iteration,
        for example enforcing positivity. The function should take one
        argument and modify it in-place.
    callback : callable, optional
        Object executing code per iteration, e.g. plotting each iterate

    See Also
    --------
    odl.solvers.iterative.iterative.landweber :
        Optimized solver for the case ``f(x) = ||Ax - b||_2^2``
    odl.solvers.iterative.iterative.conjugate_gradient :
        Optimized solver for the case ``f(x) = x^T Ax - 2 x^T b``

    References
    ----------
    [BV2004] Boyd, S, and Vandenberghe, L. *Convex optimization*.
    Cambridge university press, 2004.

    [GNS2009] Griva, I, Nash, S G, and Sofer, A. *Linear and nonlinear
    optimization*. Siam, 2009.
    """
    grad = f.gradient
    if x not in grad.domain:
        raise TypeError('`x` {!r} is not in the domain of `grad` {!r}'
                        ''.format(x, grad.domain))

    if not callable(line_search):
        line_search = ConstantLineSearch(line_search)

    grad_x = grad.range.element()
    for _ in range(maxiter):
        grad(x, out=grad_x)

        dir_derivative = -grad_x.norm() ** 2
        if abs(dir_derivative) < tol:
            return  # we have converged
        step = line_search(x, -grad_x, dir_derivative)

        x.lincomb(1, x, -step, grad_x)

        if projection is not None:
            projection(x)

        if callback is not None:
            callback(x)


def adam(f, x, learning_rate=1e-3, beta1=0.9, beta2=0.999, eps=1e-8,
         maxiter=1000, tol=1e-16, callback=None):
    r"""ADAM method to minimize an objective function.

    General implementation of ADAM for solving

    .. math::
        \min f(x)

    where :math:`f` is a differentiable functional.

    The algorithm is described in [KB2015] (`arxiv
    <https://arxiv.org/abs/1412.6980>`_). All parameter names and default
    valuesare taken from the article.

    Parameters
    ----------
    f : `Functional`
        Goal functional. Needs to have ``f.gradient``.
    x : ``f.domain`` element
        Starting point of the iteration, updated in place.
    learning_rate : positive float, optional
        Step length of the method.
    beta1 : float in [0, 1), optional
        Update rate for first order moment estimate.
    beta2 : float in [0, 1), optional
        Update rate for second order moment estimate.
    eps : positive float, optional
        A small constant for numerical stability.
    maxiter : int, optional
        Maximum number of iterations.
    tol : positive float, optional
        Tolerance that should be used for terminating the iteration.
    callback : callable, optional
        Object executing code per iteration, e.g. plotting each iterate.

    See Also
    --------
    odl.solvers.smooth.gradient.steepest_descent : Simple gradient descent.
    odl.solvers.iterative.iterative.landweber :
        Optimized solver for the case ``f(x) = ||Ax - b||_2^2``.
    odl.solvers.iterative.iterative.conjugate_gradient :
        Optimized solver for the case ``f(x) = x^T Ax - 2 x^T b``.

    References
    ----------
    [KB2015] Kingma, D P and Ba, J.
    *Adam: A Method for Stochastic Optimization*, ICLR 2015.
    """
    grad = f.gradient
    if x not in grad.domain:
        raise TypeError('`x` {!r} is not in the domain of `grad` {!r}'
                        ''.format(x, grad.domain))

    m = grad.domain.zero()
    v = grad.domain.zero()

    grad_x = grad.range.element()
    for _ in range(maxiter):
        grad(x, out=grad_x)

        if grad_x.norm() < tol:
            return

        m.lincomb(beta1, m, 1 - beta1, grad_x)
        v.lincomb(beta2, v, 1 - beta2, grad_x ** 2)

        step = learning_rate * math.sqrt(1 - beta2) / (1 - beta1)

        x.lincomb(1, x, -step, m / (sqrt(v) + eps))

        if callback is not None:
            callback(x)


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Testing utilities."""

from __future__ import absolute_import, division, print_function

import os
import sys
import warnings
from builtins import object
from contextlib import contextmanager
from time import time
from odl.core.array_API_support.comparisons import allclose, isclose, odl_all_equal
import numpy as np
import pytest

from odl.core.util.utility import is_string, run_from_ipython
from odl.core.util.dtype_utils import (
    is_boolean_dtype, is_signed_int_dtype, is_unsigned_int_dtype,
    is_floating_dtype, is_complex_dtype)

from itertools import zip_longest


skip_if_no_pytorch = pytest.mark.skipif(
        "not 'pytorch' in odl.core.space.entry_points.TENSOR_SPACE_IMPLS",
        reason='pytorch not available not available',
    )

__all__ = (
    'dtype_ndigits',
    'dtype_tol',
    'all_equal',
    'all_almost_equal',
    'is_subdict',
    'skip_if_no_pyfftw',
    'skip_if_no_pywavelets',
    'simple_fixture',
    'noise_array',
    'noise_element',
    'noise_elements',
    'fail_counter',
    'timer',
    'timeit',
    'ProgressBar',
    'ProgressRange',
    'test',
    'run_doctests',
    'test_file',
)


def _ndigits(a, b, default=None):
    """Return number of expected correct digits comparing ``a`` and ``b``.

    The returned number is the minimum `dtype_ndigits` of the two objects.

    See Also
    --------
    dtype_ndigits
    """
    dtype1 = getattr(a, 'dtype', object)
    dtype2 = getattr(b, 'dtype', object)
    return min(dtype_ndigits(dtype1, default), dtype_ndigits(dtype2, default))


def dtype_ndigits(dtype, default=None):
    """Return the number of correct digits expected for a given dtype.

    This is intended as a somewhat generous default (relative) precision for
    results of more or less stable computations.

    Returned numbers:

    - ``np.float16``: ``1``
    - ``np.float32`` or ``np.complex64``: ``3``
    - Others: ``default`` if given, otherwise ``5``

    See Also
    --------
    dtype_tol : Same precision expressed as tolerance
    """
    small_dtypes = [np.float32, np.complex64]
    tiny_dtypes = [np.float16]

    if dtype in tiny_dtypes:
        return 1
    elif dtype in small_dtypes:
        return 3
    else:
        return default if default is not None else 5


def dtype_tol(dtype, default=None):
    """Return a tolerance for a given dtype.

    This is intended as a somewhat generous default (relative) tolerance for
    results of more or less stable computations.

    Returned numbers:

    - ``np.float16``: ``1e-1``
    - ``np.float32`` or ``np.complex64``: ``1e-3``
    - Others: ``default`` if given, otherwise ``1e-5``

    See Also
    --------
    dtype_ndigits : Same tolerance expressed in number of digits.
    """
    return 10 ** -dtype_ndigits(dtype, default)


def all_equal(iter1, iter2):
    """Return ``True`` if all elements in ``a`` and ``b`` are equal.
    This is a more forgiving version of `odl_all_equal`, allowing also comparisons
    between e.g. a list and a `LinearSpaceElement` rather than requiring both sides
    to be compatible."""

    # Direct comparison for scalars, tuples or lists

    from odl.core.set.space import LinearSpaceElement

    if isinstance(iter1, LinearSpaceElement) and isinstance(iter2, LinearSpaceElement):
        return iter1 == iter2
    elif isinstance(iter1, LinearSpaceElement):
        try:
            return iter1 == iter1.space.element(iter2)
        except ValueError as e:
            pass
        except TypeError as e:
            pass
    elif isinstance(iter2, LinearSpaceElement):
        return iter2.space.element(iter1) == iter2

    try:
        if iter1 == iter2:
            return True
    except ValueError:  # Raised by NumPy when comparing arrays
        pass
    except RuntimeError:  # Raised by PyTorch when comparing tensors
        pass

    # Special case for None
    if iter1 is None and iter2 is None:
        return True

    # If one nested iterator is exhausted, go to direct comparison
    try:
        it1 = iter(iter1)
        it2 = iter(iter2)
    except TypeError:
        try:
            return iter1 == iter2
        except ValueError:  # Raised by NumPy when comparing arrays
            return False

    diff_length_sentinel = object()

    # Compare element by element and return False if the sequences have
    # different lengths
    for [ip1, ip2] in zip_longest(it1, it2,
                                  fillvalue=diff_length_sentinel):
        # Verify that none of the lists has ended (then they are not the
        # same size)
        if ip1 is diff_length_sentinel or ip2 is diff_length_sentinel:
            return False

        if not all_equal(ip1, ip2):
            return False

    return True


def all_almost_equal_array(v1, v2, ndigits):
    if v1.dtype is np.dtype(object) or v2.dtype is np.dtype(object):
        if len(v1) != len(v2):
            return False
        for w1, w2 in zip(v1, v2):
            if not all_almost_equal(w1, w2, ndigits):
                return False
        return True
    else:
        return allclose(v1, v2,
                           rtol=10 ** -ndigits, atol=10 ** -ndigits,
                           equal_nan=True)


def all_almost_equal(iter1, iter2, ndigits=None):
    """Return ``True`` if all elements in ``a`` and ``b`` are almost equal."""
    try:
        if iter1 is iter2 or all_equal(iter1, iter2):
            return True
    except ValueError:
        pass
    except RuntimeError:
        pass

    if iter1 is None and iter2 is None:
        return True

    if hasattr(iter1, '__array__') and hasattr(iter2, '__array__'):
        # Only get default ndigits if comparing arrays, need to keep `None`
        # otherwise for recursive calls.
        if ndigits is None:
            ndigits = _ndigits(iter1, iter2, None)
        return all_almost_equal_array(iter1, iter2, ndigits)

    try:
        it1 = iter(iter1)
        it2 = iter(iter2)
    except TypeError:
        if ndigits is None:
            ndigits = _ndigits(iter1, iter2, None)
        return isclose(iter1, iter2,
                          atol=10 ** -ndigits, rtol=10 ** -ndigits,
                          equal_nan=True)

    diff_length_sentinel = object()
    for [ip1, ip2] in zip_longest(it1, it2,
                                  fillvalue=diff_length_sentinel):
        # Verify that none of the lists has ended (then they are not the
        # same size)
        if ip1 is diff_length_sentinel or ip2 is diff_length_sentinel:
            return False

        if not all_almost_equal(ip1, ip2, ndigits):
            return False

    return True


def is_subdict(subdict, dictionary):
    """Return ``True`` if all items of ``subdict`` are in ``dictionary``."""
    return all(item in dictionary.items() for item in subdict.items())


try:
    import pytest

except ImportError:

    def identity(*args, **kwargs):
        if args and callable(args[0]):
            return args[0]
        else:
            return identity

    skip_if_no_pyfftw = identity
    skip_if_no_pywavelets = identity

else:
    # Mark decorators for test parameters
    skip_if_no_pyfftw = pytest.mark.skipif(
        'not odl.trafos.PYFFTW_AVAILABLE',
        reason='pyFFTW not available',
    )
    skip_if_no_pywavelets = pytest.mark.skipif(
        'not odl.trafos.PYWT_AVAILABLE',
        reason='PyWavelets not available',
    )


def simple_fixture(name, params, fmt=None):
    """Helper to create a pytest fixture using only name and params.

    Parameters
    ----------
    name : str
        Name of the parameters used for the ``ids`` argument
        to `pytest.fixture`.
    params : sequence
        Values to be taken as parameters in the fixture. They are
        used as ``params`` argument to `_pytest.fixtures.fixture`.
        Arguments wrapped in a ``pytest.skipif`` decorator are
        unwrapped for the generation of the test IDs.
    fmt : str, optional
        Use this format string for the generation of the ``ids``.
        For each value, the id string is generated as ::

            fmt.format(name=name, value=value)

        hence the format string must use ``{name}`` and ``{value}``.
        Default format strings are:

            - ``" {name}='{value}' "`` for string parameters,
            - ``" {name}={value} "`` for other types.
    """
    import _pytest

    if fmt is None:
        # Use some intelligence to make good format strings
        fmt_str = " {name}='{value}' "
        fmt_default = " {name}={value} "

        ids = []
        for p in params:
            # TODO: other types of decorators?
            if (
                isinstance(p, _pytest.mark.MarkDecorator)
                and p.name == 'skipif'
            ):
                # Unwrap the wrapped object in the decorator
                if is_string(p.args[1]):
                    ids.append(fmt_str.format(name=name, value=p.args[1]))
                else:
                    ids.append(fmt_default.format(name=name, value=p.args[1]))
            else:
                if is_string(p):
                    ids.append(fmt_str.format(name=name, value=p))
                else:
                    ids.append(fmt_default.format(name=name, value=p))
    else:
        # Use provided `fmt` for everything
        ids = [fmt.format(name=name, value=p) for p in params]

    wrapper = pytest.fixture(scope='module', ids=ids, params=params)
    return wrapper(lambda request: request.param)


# Helpers to generate data
def noise_array(space):
    """Generate a white noise array that is compatible with ``space``.

    The array contains white noise with standard deviation 1 in the case of
    floating point dtypes and uniformly spaced values between -10 and 10 in
    the case of integer dtypes.

    For product spaces the method is called recursively for all sub-spaces.

    Notes
    -----
    This method is intended for internal testing purposes. For more explicit
    example elements see ``odl.core.phantoms`` and ``LinearSpaceElement.examples``.

    Parameters
    ----------
    space : `LinearSpace`
        Space from which to derive the array data type and size.

    Returns
    -------
    noise_array : `numpy.ndarray` element
        Array with white noise such that ``space.element``'s can be created
        from it.

    Examples
    --------
    Create single noise array:

    >>> space = odl.rn(3)
    >>> array = noise_array(space)

    See Also
    --------
    noise_element
    noise_elements
    odl.core.set.space.LinearSpace.examples : Examples of elements
        typical to the space.
    """
    from odl.core.space import ProductSpace
    if isinstance(space, ProductSpace):

        if space.is_power_space:
            return [noise_array(si) for si in space]

        # Non-power–product-space elements are represented as arrays of arrays,
        # each in general with a different shape. These cannot be monolithic
        # NumPy arrays. NumPy allows non-rectangular arrays when explicitly
        # requesting dtype=object, but these behave different from ordinary
        # arrays in several ways. The following is a hack to have only the
        # outer array with dtype=object but store the inner elements as for the
        # constituent spaces. The resulting ragged arrays support some, but not
        # all numerical operations.
        result = [None for si in space]
        for i, si in enumerate(space):
            result[i] = noise_array(si)
        return result

    else:
        dtype = space.dtype_identifier
        if is_boolean_dtype(dtype):
            arr = np.random.randint(0, 2, size=space.shape, dtype=dtype)
        elif is_unsigned_int_dtype(dtype):
            arr = np.random.randint(0, 10, space.shape)
        elif is_signed_int_dtype(dtype):
            arr = np.random.randint(-10, 10, space.shape)
        elif is_floating_dtype(dtype):
            arr = np.random.randn(*space.shape)
        elif is_complex_dtype(dtype):
            arr = (
                np.random.randn(*space.shape)
                + 1j * np.random.randn(*space.shape)
            ) / np.sqrt(2.0)
        else:
            raise ValueError('bad dtype {}'.format(space.dtype))
        return space.element(arr).data


def noise_element(space):
    """Create a white noise element in ``space``.

    The element contains white noise with standard deviation 1 in the case of
    floating point dtypes and uniformly spaced values between -10 and 10 in
    the case of integer dtypes.

    For product spaces the method is called recursively for all sub-spaces.

    Notes
    -----
    This method is intended for internal testing purposes. For more explicit
    example elements see ``odl.core.phantoms`` and ``LinearSpaceElement.examples``.

    Parameters
    ----------
    space : `LinearSpace`
        Space in which to create an element. The
        `odl.core.set.space.LinearSpace.element` method of the space needs to
        accept input of `numpy.ndarray` type.

    Returns
    -------
    noise_element : ``space`` element

    Examples
    --------
    Create single noise element:

    >>> space = odl.rn(3)
    >>> vector = noise_element(space)

    See Also
    --------
    noise_array
    noise_elements
    odl.core.set.space.LinearSpace.examples : Examples of elements typical
        to the space.
    """
    return space.element(noise_array(space))


def noise_elements(space, n=1):
    """Create a list of ``n`` noise arrays and elements in ``space``.

    The arrays contain white noise with standard deviation 1 in the case of
    floating point dtypes and uniformly spaced values between -10 and 10 in
    the case of integer dtypes.

    The returned elements have the same values as the arrays.

    For product spaces the method is called recursively for all sub-spaces.

    Notes
    -----
    This method is intended for internal testing purposes. For more explicit
    example elements see ``odl.core.phantoms`` and ``LinearSpaceElement.examples``.

    Parameters
    ----------
    space : `LinearSpace`
        Space in which to create an element. The
        `odl.core.set.space.LinearSpace.element` method of the space needs to
        accept input of `numpy.ndarray` type.
    n : int, optional
        Number of elements to create.

    Returns
    -------
    arrays : `numpy.ndarray` or tuple of `numpy.ndarray`
        A single array if ``n == 1``, otherwise a tuple of arrays.
    elements : ``space`` element or tuple of ``space`` elements
        A single element if ``n == 1``, otherwise a tuple of elements.

    Examples
    --------
    Create single noise element:

    >>> space = odl.rn(3)
    >>> arr, vector = noise_elements(space)

    Create multiple noise elements:

    >>> [arr1, arr2], [vector1, vector2] = noise_elements(space, n=2)

    See Also
    --------
    noise_array
    noise_element
    """
    arrs = tuple(noise_array(space) for _ in range(n))

    # Make space elements from arrays
    elems = tuple(space.element(arr) for arr in arrs)

    if n == 1:
        return tuple(arrs + elems)
    else:
        return arrs, elems


@contextmanager
def fail_counter(test_name, err_msg=None, logger=print):
    """Used to count the number of failures of something.

    Usage::

        with fail_counter("my_test") as counter:
            # Do stuff

            counter.fail()

    When done, it prints ::

        my_test
        *** FAILED 1 TEST CASE(S) ***
    """

    class _FailCounter(object):

        def __init__(self):
            self.num_failed = 0
            self.fail_strings = []

        def fail(self, string=None):
            """Add failure with reason as string."""
            # TODO: possibly limit number of printed strings
            self.num_failed += 1
            if string is not None:
                self.fail_strings.append(str(string))

    try:
        counter = _FailCounter()
        yield counter
    finally:

        if counter.num_failed == 0:
            logger('{:<70}: Completed all test cases.'.format(test_name))
        else:
            print(test_name)
            for fail_string in counter.fail_strings:
                print(fail_string)
            if err_msg is not None:
                print(err_msg)
            print('*** FAILED {} TEST CASE(S) ***'.format(counter.num_failed))


@contextmanager
def timer(name=None):
    """A timer context manager.

    Usage::

        with timer('name'):
            # Do stuff

    Prints the time stuff took to execute.
    """
    if name is None:
        name = "Elapsed"

    try:
        tstart = time()
        yield
    finally:
        time_str = '{:.3f}'.format(time() - tstart)
        print('{:>30s} : {:>10s} '.format(name, time_str))


def timeit(arg):
    """A timer decorator.

    Usage::

        @timeit
        def myfunction(...):
            ...

        @timeit('info string')
        def myfunction(...):
            ...
    """

    if callable(arg):
        def timed_function(*args, **kwargs):
            with timer(str(arg)):
                return arg(*args, **kwargs)
        return timed_function
    else:
        def _timeit_helper(func):
            def timed_function(*args, **kwargs):
                with timer(arg):
                    return func(*args, **kwargs)
            return timed_function
        return _timeit_helper


class ProgressBar(object):

    """A simple command-line progress bar.

    Usage:

    >>> progress = ProgressBar('Reading data', 10)
    \rReading data: [                              ] Starting
    >>> progress.update(4) #halfway, zero indexing
    \rReading data: [###############               ] 50.0%

    Multi-indices, from slowest to fastest:

    >>> progress = ProgressBar('Reading data', 10, 10)
    \rReading data: [                              ] Starting
    >>> progress.update(9, 8)
    \rReading data: [############################# ] 99.0%

    Supports simply calling update, which moves the counter forward:

    >>> progress = ProgressBar('Reading data', 10, 10)
    \rReading data: [                              ] Starting
    >>> progress.update()
    \rReading data: [                              ]  1.0%
    """

    def __init__(self, text='progress', *njobs):
        """Initialize a new instance."""
        self.text = str(text)
        if len(njobs) == 0:
            raise ValueError('need to provide at least one job')
        self.njobs = njobs
        self.current_progress = 0.0
        self.index = 0
        self.done = False
        self.start()

    def start(self):
        """Print the initial bar."""
        sys.stdout.write('\r{0}: [{1:30s}] Starting'.format(self.text,
                                                            ' ' * 30))

        sys.stdout.flush()

    def update(self, *indices):
        """Update the bar according to ``indices``."""
        if indices:
            if len(indices) != len(self.njobs):
                raise ValueError('number of indices not correct')
            self.index = np.ravel_multi_index(indices, self.njobs) + 1
        else:
            self.index += 1

        # Find progress as ratio between 0 and 1
        # offset by 1 for zero indexing
        progress = self.index / np.prod(self.njobs)

        # Write a progressbar and percent
        if progress < 1.0:
            # Only update on 0.1% intervals
            if progress > self.current_progress + 0.001:
                sys.stdout.write('\r{0}: [{1:30s}] {2:4.1f}%   '.format(
                    self.text, '#' * int(30 * progress), 100 * progress))
                self.current_progress = progress
        else:  # Special message when done
            if not self.done:
                sys.stdout.write('\r{0}: [{1:30s}] Done      \n'.format(
                    self.text, '#' * 30))
                self.done = True

        sys.stdout.flush()


class ProgressRange(object):

    """Simple range sequence with progress bar output"""

    def __init__(self, text, n):
        """Initialize a new instance."""
        self.current = 0
        self.n = n
        self.bar = ProgressBar(text, n)

    def __iter__(self):
        return self

    def __next__(self):
        if self.current < self.n:
            val = self.current
            self.current += 1
            self.bar.update()
            return val
        else:
            raise StopIteration()


def test(arguments=None):
    """Run ODL tests given by arguments."""
    try:
        import pytest
    except ImportError:
        raise ImportError(
            'ODL tests cannot be run without `pytest` installed.\n'
            'Run `$ pip install [--user] odl[testing]` in order to install '
            '`pytest`.'
        )

    from .pytest_config import collect_ignore

    this_dir = os.path.dirname(__file__)
    odl_root = os.path.abspath(os.path.join(this_dir, os.pardir, os.pardir, os.pardir))

    args = ['{root}/odl/test'.format(root=odl_root)]

    ignores = ['--ignore={}'.format(file) for file in collect_ignore]
    args.extend(ignores)

    if arguments is not None:
        args.extend(arguments)

    pytest.main(args)


def run_doctests(skip_if=False, **kwargs):
    """Run all doctests in the current module.

    This function calls ``doctest.testmod()``, by default with the options
    ``optionflags=doctest.NORMALIZE_WHITESPACE`` and
    ``extraglobs={'odl': odl, 'np': np}``. This can be changed with
    keyword arguments.

    Parameters
    ----------
    skip_if : bool
        For ``True``, skip the doctests in this module.
    kwargs :
        Extra keyword arguments passed on to the ``doctest.testmod``
        function.
    """
    from doctest import testmod, NORMALIZE_WHITESPACE, SKIP
    from packaging.version import parse as parse_version
    import odl
    import numpy as np

    optionflags = kwargs.pop('optionflags', NORMALIZE_WHITESPACE)
    if skip_if:
        optionflags |= SKIP

    extraglobs = kwargs.pop('extraglobs', {'odl': odl, 'np': np})

    if run_from_ipython():
        try:
            import spyder
        except ImportError:
            pass
        else:
            if parse_version(spyder.__version__) < parse_version('3.1.4'):
                warnings.warn('A bug with IPython and Spyder < 3.1.4 '
                              'sometimes causes doctests to fail to run. '
                              'Please upgrade Spyder or use another '
                              'interpreter if the doctests do not work.',
                              RuntimeWarning)

    testmod(optionflags=optionflags, extraglobs=extraglobs, **kwargs)


def test_file(file, args=None):
    """Run tests in file with proper default arguments."""
    try:
        import pytest
    except ImportError:
        raise ImportError('ODL tests cannot be run without `pytest` installed.'
                          '\nRun `$ pip install [--user] odl[testing]` in '
                          'order to install `pytest`.')

    if args is None:
        args = []

    args.extend([str(file.replace('\\', '/')), '-v', '--capture=sys'])

    pytest.main(args)


# What types will auto-chosen in expressions like `odl.vector([1.0,2.0])` is
# backend-dependent, with NumPy prioritizing precision and PyTorch speed. This
# follows what the underlying `np.array` / `torch.tensor` constructors choose.
# Note that this differs from what happens when `float` is explicitly specified
# as the `dtype` - this will always be interpreted as double precision
# (see `DTYPE_SHORTHANDS`).
default_precision_dict = {
    'pytorch':{
        'integer' : 'int32',
        'float'   : 'float32',
        'complex' : 'complex64'
    },
    'numpy':{
        'integer' : 'int64',
        'float'   : 'float64',
        'complex' : 'complex128'
    }
}

if __name__ == '__main__':
    run_doctests()



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Maximum Likelihood Expectation Maximization algorithm."""

from __future__ import print_function, division, absolute_import
import numpy as np
from odl.core.array_API_support import maximum, any, log, sum

__all__ = ('mlem', 'osmlem', 'poisson_log_likelihood')


def mlem(op, x, data, niter, callback=None, **kwargs):

    r"""Maximum Likelihood Expectation Maximation algorithm.

    Attempts to solve::

        max_x L(x | data)

    where ``L(x | data)`` is the Poisson likelihood of ``x`` given ``data``.
    The likelihood depends on the forward operator ``op`` such that
    (approximately)::

        op(x) = data


    Parameters
    ----------
    op : `Operator`
        Forward operator in the inverse problem.
    x : ``op.domain`` element
        Vector to which the result is written. Its initial value is
        used as starting point of the iteration, and its values are
        updated in each iteration step.
        The initial value of ``x`` should be non-negative.
    data : ``op.range`` `element-like`
        Right-hand side of the equation defining the inverse problem.
    niter : int
        Number of iterations.
    callback : callable, optional
        Function called with the current iterate after each iteration.

    Other Parameters
    ----------------
    sensitivities : float or ``op.domain`` `element-like`, optional
        The algorithm contains a ``A^T 1``
        term, if this parameter is given, it is replaced by it.
        Default: ``op.adjoint(op.range.one())``

    Notes
    -----
    Given a forward model :math:`A` and data :math:`g`,
    the algorithm attempts to find an :math:`x` that maximizes:

    .. math::
        P(g | g \text{ is } Poisson(A(x)) \text{ distributed}).

    The algorithm is explicitly given by:

    .. math::
       x_{n+1} = \frac{x_n}{A^* 1} A^* (g / A(x_n))

    See Also
    --------
    osmlem : Ordered subsets MLEM
    loglikelihood : Function for calculating the logarithm of the likelihood
    """
    osmlem([op], x, [data], niter=niter, callback=callback,
           **kwargs)


def osmlem(op, x, data, niter, callback=None, **kwargs):
    r"""Ordered Subsets Maximum Likelihood Expectation Maximation algorithm.

    This solver attempts to solve::

        max_x L(x | data)

    where ``L(x, | data)`` is the likelihood of ``x`` given ``data``. The
    likelihood depends on the forward operators ``op[0], ..., op[n-1]`` such
    that (approximately)::

        op[i](x) = data[i]


    Parameters
    ----------
    op : sequence of `Operator`
        Forward operators in the inverse problem.
    x : ``op.domain`` element
        Vector to which the result is written. Its initial value is
        used as starting point of the iteration, and its values are
        updated in each iteration step.
        The initial value of ``x`` should be non-negative.
      data : sequence of ``op.range`` `element-like`
        Right-hand sides of the equation defining the inverse problem.
    niter : int
        Number of iterations.
    callback : callable, optional
        Function called with the current iterate after each iteration.

    Other Parameters
    ----------------
    sensitivities : float or ``op.domain`` `element-like`, optional
        The algorithm contains an ``A^T 1``
        term, if this parameter is given, it is replaced by it.
        Default: ``op[i].adjoint(op[i].range.one())``

    Notes
    -----
    Given forward models :math:`A_i`, and data :math:`g_i`,
    :math:`i = 1, ..., M`,
    the algorithm attempts to find an :math:`x` that
    maximizes:

    .. math::
        \prod_{i=1}^M P(g_i | g_i \text{ is }
        Poisson(A_i(x)) \text{ distributed}).

    The algorithm is explicitly given by partial updates:

    .. math::
       x_{n + m/M} =
       \frac{x_{n + (m - 1)/M}}{A_i^* 1} A_i^* (g_i / A_i(x_{n + (m - 1)/M}))

    for :math:`m = 1, ..., M` and :math:`x_{n+1} = x_{n + M/M}`.

    The algorithm is not guaranteed to converge, but works for many practical
    problems.

    References
    ----------
    Natterer, F. Mathematical Methods in Image Reconstruction, section 5.3.2.

    See Also
    --------
    mlem : Ordinary MLEM algorithm without subsets.
    loglikelihood : Function for calculating the logarithm of the likelihood
    """
    n_ops = len(op)
    if len(data) != n_ops:
        raise ValueError('number of data ({}) does not match number of '
                         'operators ({})'.format(len(data), n_ops))
    if not all(x in opi.domain for opi in op):
        raise ValueError('`x` not an element in the domains of all operators')

    # Convert data to range elements
    data = [op[i].range.element(data[i]) for i in range(len(op))]

    # Parameter used to enforce positivity.
    # TODO: let users give this.
    eps = 1e-8

    if any(x < 0):
        raise ValueError('`x` must be non-negative')

    # Extract the sensitivites parameter
    sensitivities = kwargs.pop('sensitivities', None)
    if sensitivities is None:
        sensitivities = [maximum(opi.adjoint(opi.range.one()), eps)
                         for opi in op]
    else:
        # Make sure the sensitivities is a list of the correct size.
        try:
            list(sensitivities)
        except TypeError:
            sensitivities = [sensitivities] * n_ops

    tmp_dom = op[0].domain.element()
    tmp_ran = [opi.range.element() for opi in op]

    for _ in range(niter):
        for i in range(n_ops):
            op[i](x, out=tmp_ran[i])
            maximum(tmp_ran[i], eps, out=tmp_ran[i])

            data[i].divide(tmp_ran[i], out=tmp_ran[i])

            op[i].adjoint(tmp_ran[i], out=tmp_dom)
            tmp_dom /= sensitivities[i]

            x *= tmp_dom

            if callback is not None:
                callback(x)


def poisson_log_likelihood(x, data):
    """Poisson log-likelihood of ``data`` given noise parametrized by ``x``.

    Parameters
    ----------
    x : ``op.domain`` element
        Value to condition the log-likelihood on.
    data : ``op.range`` element
        Data whose log-likelihood given ``x`` shall be calculated.
    """
    if any(x < 0):
        raise ValueError('`x` must be non-negative')

    return sum(data * log(x + 1e-8) - x)



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Test skimage back-end."""

from __future__ import division
import numpy as np

import odl
from odl.applications.tomo.backends.skimage_radon import (
    skimage_radon_forward_projector, skimage_radon_back_projector)
from odl.applications.tomo.util.testutils import skip_if_no_skimage


@skip_if_no_skimage
def test_skimage_radon_projector_parallel2d():
    """Parallel 2D forward and backward projectors with skimage."""

    # Create reco space and a phantom
    reco_space = odl.uniform_discr([-5, -5], [5, 5], (5, 5))
    phantom = odl.core.phantom.cuboid(reco_space, min_pt=[0, 0], max_pt=[5, 5])

    # Create parallel geometry
    angle_part = odl.uniform_partition(0, np.pi, 5)
    det_part = odl.uniform_partition(-6, 6, 6)
    geom = odl.applications.tomo.Parallel2dGeometry(angle_part, det_part)

    # Make projection space
    proj_space = odl.uniform_discr_frompartition(geom.partition)

    # Forward evaluation
    proj_data = skimage_radon_forward_projector(phantom, geom, proj_space)
    assert proj_data.shape == proj_space.shape
    assert proj_data.norm() > 0

    # Backward evaluation
    backproj = skimage_radon_back_projector(proj_data, geom, reco_space)
    assert backproj.shape == reco_space.shape
    assert backproj.norm() > 0


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Bindings to the PyWavelets backend for wavelet transforms.

`PyWavelets <https://pywavelets.readthedocs.io/>`_ is a Python library
for wavelet transforms in arbitrary dimensions, featuring a large number
of built-in wavelet filters.
"""

from __future__ import print_function, division, absolute_import

import numpy as np

try:
    import pywt
    PYWT_AVAILABLE = True
except ImportError:
    PYWT_AVAILABLE = False


__all__ = ('PAD_MODES_ODL2PYWT', 'PYWT_SUPPORTED_MODES', 'PYWT_AVAILABLE',
           'pywt_wavelet', 'pywt_pad_mode', 'precompute_raveled_slices')


# A clear illustration of all of these padding modes is available at:
# https://pywavelets.readthedocs.io/en/latest/ref/signal-extension-modes.html
PAD_MODES_ODL2PYWT = {'constant': 'zero',
                      'periodic': 'periodic',
                      'symmetric': 'symmetric',
                      'order0': 'constant',
                      'order1': 'smooth',
                      'pywt_periodic': 'periodization',
                      'reflect': 'reflect',
                      'antireflect': 'antireflect',
                      'antisymmetric': 'antisymmetric',
                      }
PYWT_SUPPORTED_MODES = PAD_MODES_ODL2PYWT.values()


def pywt_wavelet(wavelet):
    """Convert ``wavelet`` to a `pywt.Wavelet` instance."""
    if isinstance(wavelet, pywt.Wavelet):
        return wavelet
    else:
        return pywt.Wavelet(wavelet)


def pywt_pad_mode(pad_mode, pad_const=0):
    """Convert ODL-style padding mode to pywt-style padding mode.

    Parameters
    ----------
    pad_mode : str
        The ODL padding mode to use at the boundaries.
    pad_const : float, optional
        Value to use outside the signal boundaries when ``pad_mode`` is
        'constant'. Only a value of 0. is supported by PyWavelets

    Returns
    -------
    pad_mode_pywt : str
        The corresponding name of the requested padding mode in PyWavelets.
        See `signal extension modes`_.

    References
    ----------
    .. _signal extension modes:
       https://pywavelets.readthedocs.io/en/latest/ref/signal-extension-modes.html
    """
    pad_mode = str(pad_mode).lower()
    if pad_mode == 'constant' and pad_const != 0.0:
        raise ValueError('constant padding with constant != 0 not supported '
                         'for `pywt` back-end')
    try:
        return PAD_MODES_ODL2PYWT[pad_mode]
    except KeyError:
        raise ValueError("`pad_mode` '{}' not understood".format(pad_mode))


def precompute_raveled_slices(coeff_shapes, axes=None):
    """Return slices and shapes for raveled multilevel wavelet coefficients.

    The output is equivalent to the ``coeff_slices`` output of
    `pywt.ravel_coeffs`, but this function does not require computing a
    wavelet transform first.

    Parameters
    ----------
    coeff_shapes : array-like
        A list of multilevel wavelet coefficient shapes as returned by
        `pywt.wavedecn_shapes`.
    axes : sequence of ints, optional
        Axes over which the DWT that created ``coeffs`` was performed. The
        default value of None corresponds to all axes.

    Returns
    -------
    coeff_slices : list
        List of slices corresponding to each coefficient. As a 2D example,
        ``coeff_arr[coeff_slices[1]['dd']]`` would extract the first level
        detail coefficients from ``coeff_arr``.

    Examples
    --------
    >>> import pywt
    >>> data_shape = (64, 64)
    >>> coeff_shapes = pywt.wavedecn_shapes(data_shape, wavelet='db2', level=3,
    ...                                     mode='periodization')
    >>> coeff_slices = precompute_raveled_slices(coeff_shapes)
    >>> print(coeff_slices[0])  # approximation coefficients
    slice(None, 64, None)
    >>> d1_coeffs = coeff_slices[-1]  # first level detail coefficients
    >>> (d1_coeffs['ad'], d1_coeffs['da'], d1_coeffs['dd'])
    (slice(1024, 2048, None), slice(2048, 3072, None), slice(3072, 4096, None))
    """
    # initialize with the approximation coefficients.
    a_shape = coeff_shapes[0]
    a_size = np.prod(a_shape)

    if len(coeff_shapes) == 1:
        # only a single approximation coefficient array was found
        return [slice(a_size), ]

    a_slice = slice(a_size)

    # initialize list of coefficient slices
    coeff_slices = []
    coeff_slices.append(a_slice)

    # loop over the detail cofficients, embedding them in coeff_arr
    details_list = coeff_shapes[1:]
    offset = a_size
    for shape_dict in details_list:
        # new dictionaries for detail coefficient slices and shapes
        coeff_slices.append({})
        keys = sorted(shape_dict.keys())
        for key in keys:
            shape = shape_dict[key]
            size = np.prod(shape)
            sl = slice(offset, offset + size)
            offset += size
            coeff_slices[-1][key] = sl
    return coeff_slices


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests(skip_if=not PYWT_AVAILABLE)



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import division
import pytest
try:
    import pywt
except ImportError:
    pass

import odl
from odl.trafos.backends.pywt_bindings import (
    PYWT_AVAILABLE, PAD_MODES_ODL2PYWT, pywt_wavelet, pywt_pad_mode)
from odl.core.util.testutils import (simple_fixture)

pytestmark = pytest.mark.skipif(not PYWT_AVAILABLE,
                                reason='`pywt` backend not available')


# --- pytest fixtures --- #
wavelet = simple_fixture('wavelet_name', ['db1', 'sym2'])
odl_mode = simple_fixture('odl_mode', ['constant', 'order0', 'order1'])


def test_pywt_wavelet(wavelet):
    # pywt_wavelet takes either string or pywt.Wavelet object as input
    wavelet = pywt_wavelet(wavelet)
    assert isinstance(wavelet, pywt.Wavelet)

    wavelet2 = pywt_wavelet(wavelet)
    assert isinstance(wavelet2, pywt.Wavelet)
    assert wavelet2 is wavelet


def test_pywt_pad_mode(odl_mode):
    pywt_mode = pywt_pad_mode(odl_mode)
    assert pywt_mode in list(PAD_MODES_ODL2PYWT.values())


def test_pywt_pad_errors():
    with pytest.raises(ValueError):
        pywt_pad_mode('constant', pad_const=1.)
    with pytest.raises(ValueError):
        pywt_pad_mode('invalid mode')


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2025 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""
Array creation functions expected by the python array API.
Although ODL has many ways to create a tensor, we have found useful during development and testing to be able to create arrays in a certain backend.
We do not expect the users to work with these functions often but have still implemented them as we deemed useful during development. 

Notes:
    -> the functions with name *_like take an array/ODL object as an input
    -> the other functions require impl, shape, dtype, device arguments.

Examples:
>>> odl.arange('numpy', 0,10,1, dtype='float32', device='cuda:0')
Traceback (most recent call last):
ValueError: Unsupported device for NumPy: 'cuda:0'
>>> odl.arange('numpy',start=0,stop=10,step=1, dtype='float32', device='cpu')
array([ 0.,  1.,  2.,  3.,  4.,  5.,  6.,  7.,  8.,  9.], dtype=float32)
>>> odl.asarray(odl.rn(4).element([1,2,3,4]))
array([ 1.,  2.,  3.,  4.])
>>> odl.full('numpy', (4,4), 4) == np.full((4,4),4)
array([[ True,  True,  True,  True],
       [ True,  True,  True,  True],
       [ True,  True,  True,  True],
       [ True,  True,  True,  True]], dtype=bool)
>>> odl.full_like(x = np.full((4,4),4), fill_value=4) == np.full((4,4),4)
array([[ True,  True,  True,  True],
       [ True,  True,  True,  True],
       [ True,  True,  True,  True],
       [ True,  True,  True,  True]], dtype=bool)
"""

from .utils import get_array_and_backend, lookup_array_backend

__all__ = (
    'arange',
    'asarray',
    'empty',
    'empty_like',
    'eye',
    # 'from_dlpack',
    'full',
    'full_like',
    'linspace',
    'meshgrid',
    'ones',
    'ones_like',
    'tril',
    'triu',
    'zeros',
    'zeros_like'
)

def _helper_from_impl(fname, impl, *args, **kwargs):
    backend = lookup_array_backend(impl)
    fn = getattr(backend.array_namespace, fname)
    return fn(*args, **kwargs)

def _helper_from_array(fname, x, **kwargs):    
    x, backend_x = get_array_and_backend(x)
    fn = getattr(backend_x.array_namespace, fname)
    return fn(x, **kwargs)

def arange(impl, start, stop=None, step=1, dtype=None, device=None):
    """
    Returns evenly spaced values within the half-open interval [start, stop) as a one-dimensional array.
    """
    return _helper_from_impl('arange', impl, start, stop=stop, step=step, dtype=dtype, device=device)

def asarray(x):
    """
    Returns an array corresponding to an ODL object.
    """
    return _helper_from_array('asarray', x)

def empty(impl, shape, dtype=None, device=None):
    """
    Returns an uninitialized array having a specified shape.
    """
    return _helper_from_impl('empty', impl, shape, dtype=dtype, device=device)

def empty_like(x, dtype=None, device=None):
    """
    Returns an uninitialized array with the same shape as an input array x.
    """
    return _helper_from_array('empty_like', x=x, dtype=dtype, device=device)

def eye(impl, n_rows, n_cols=None, k=0, dtype=None, device=None):
    """
    Returns a two-dimensional array with ones on the kth diagonal and zeros elsewhere.
    """
    return _helper_from_impl('eye', impl, n_rows=n_rows, n_cols=n_cols, k=k, dtype=dtype, device=device)

# def from_dlpack(x, device=None):
#     """
#     Returns a new array containing the data from another (array) object with a __dlpack__ method.
#     Note:
#         The device argument is currently NOT used, this is due to Pytorch needing to catch up with the array API standard
#     """
#     return _helper_from_array('from_dlpack', x=x)

def full(impl, shape, fill_value, dtype=None, device=None):
    """
    Returns a new array having a specified shape and filled with fill_value.
    """
    return _helper_from_impl('full', impl, shape=shape, fill_value=fill_value, dtype=dtype, device=device)

def full_like(x, fill_value, dtype=None, device=None):
    """
    Returns a new array filled with fill_value and having the same shape as an input array x.
    """
    return _helper_from_array('full_like', x=x, fill_value=fill_value, dtype=dtype, device=device)

def linspace(impl, start, stop, num, dtype=None, device=None, endpoint=True):
    """
    Returns evenly spaced numbers over a specified interval.
    """
    return _helper_from_impl('linspace', impl, start, stop, num, dtype=dtype, device=device, endpoint=endpoint)

def meshgrid(impl, *arrays, indexing='xy'):
    """    	
    Returns coordinate matrices from coordinate vectors.
    """
    return _helper_from_impl('meshgrid', impl, *arrays, indexing=indexing)

def ones(impl, shape, dtype=None, device=None):
    """
    Returns a new array having a specified shape and filled with ones.
    """
    return _helper_from_impl('ones', impl, shape=shape, dtype=dtype, device=device)

def ones_like(x, dtype=None, device=None):
    """
    Returns a new array filled with ones and having the same shape as an input array x.
    """
    return _helper_from_array('ones_like', x, dtype=dtype, device=device)

def tril(x, k=0):
    """
    Returns the lower triangular part of a matrix (or a stack of matrices) x.
    """
    return _helper_from_array('tril', x, k=k)

def triu(x, k=0):
    """
    Returns the upper triangular part of a matrix (or a stack of matrices) x.
    """
    return _helper_from_array('triu', x, k=k)

def zeros(impl, shape, dtype=None, device=None):
    """
    Returns a new array having a specified shape and filled with zeros.
    """
    return _helper_from_impl('zeros', impl, shape=shape, dtype=dtype, device=device)

def zeros_like(x, dtype=None, device=None):
    """
    Returns a new array filled with zeros and having the same shape as an input array x.
    """
    return _helper_from_array('zeros_like', x, dtype=dtype, device=device)


from odl.core.array_API_support import get_array_and_backend, lookup_array_backend
from .weighting import ConstWeighting, ArrayWeighting, CustomInner, CustomNorm, CustomDist

def space_weighting(
        impl : str,
        device = 'cpu',
        **kwargs
    ):
    """
    Notes: 
        To instantiate a weigthing, one can use a variety of mutually exclusive parameters
        1) inner (callable): the inner product between two elements of the space
        2) norm (callable): the norm of an element of the space
            -> sqrt(inner(x,x).real)
        3) dist (callable): the distance between two elements of the space
            -> norm(x1-x2)
        4) weight (float | ArrayLike): Scalar or element-wise weighting of the space elements
        
        In case a weight was provided, additionally the following is supported:
        4A) exponent (float): exponent of the summands in the norm, used for Banach spaces like L¹
        If the exponent is 2, the weight is then used for defining an inner product and the
        other operations, whereas for other exponents only the norm and distance are enabled.

        For a custom inner-product space, the exponent must be 2 (the default). The inner product
        also implies a norm and distance then.
        A custom norm defines a distance but will disable the inner product. A custom distance
        disables all other operations.
    """

    if 'exponent' in kwargs:
        # Pop the kwarg
        exponent = kwargs['exponent']
        assert not set(['norm', 'dist']).issubset(kwargs)
        # Assign the attribute
        if exponent != 2:
            assert 'inner' not in kwargs
    else:
        exponent = 2

    if 'inner' in kwargs:
        # Pop the kwarg
        inner = kwargs.pop('inner')
        # check the kwarg
        assert callable(inner)
        # Check the consistency
        assert exponent == 2

        for arg in ['norm', 'dist', 'weight']:
            if arg in kwargs:
                raise ValueError(f"If a custom inner product is specified, the weighting cannot also have custom {arg}={kwargs[arg]}.")
        
        return CustomInner(inner, device=device, impl=impl)
        
    elif 'norm' in kwargs:
        # Pop the kwarg
        array_norm = kwargs.pop('norm')
        # check the kwarg
        assert callable(array_norm)
        # Check the consistency
        for arg in ['exponent', 'inner', 'dist', 'weight']:
            if arg in kwargs:
                raise ValueError(f"If a custom norm is specified, the weighting cannot also have custom {arg}={kwargs[arg]}.")

        return CustomNorm(array_norm, device=device, impl=impl)
    
    elif 'dist' in kwargs:
        # Pop the kwarg
        dist  = kwargs.pop('dist')
        # check the kwarg
        assert callable(dist)
        # Check the consistency
        for arg in ['exponent', 'inner', 'norm', 'weight']:
            if arg in kwargs:
                raise ValueError(f"If a custom distance is specified, the weighting cannot also have custom {arg}={kwargs[arg]}.")
        
        
        return CustomDist(dist, device=device, impl=impl)
    
    elif 'weight' in kwargs:
        # Pop the kwarg
        weight = kwargs.pop('weight')
        # Check the consistency
        for arg in ['inner', 'norm', 'dist']:
            if arg in kwargs:
                raise ValueError(f"If a custom weight is specified, the weighting cannot also have custom {arg}={kwargs[arg]}.")
        
        if isinstance(weight, (int, float)):
            if 0 < weight and weight != float('inf'):
                weight = float(weight)
            else:
                raise ValueError("If the weight is a scalar, it must be positive")
            return ConstWeighting(const=weight, impl=impl, device=device, exponent=exponent)
        
        elif isinstance(weight, (tuple, list)):
            array_backend = lookup_array_backend(impl)
            weight = array_backend.array_constructor(weight, device=device)
            if array_backend.array_namespace.any(weight < 0):
                raise ValueError("If the weight is an array, all its elements must be positive")
                    
        elif hasattr(weight, '__array__'):
            weight, backend = get_array_and_backend(weight)
            if backend.array_namespace.all(0 < weight):
                assert device == weight.device.__str__(), f"The weighing is expecting the device {device}, but the array provided for the weight has a device {weight.device}. Please make sure that the two devices are consistent"
            else:
                raise ValueError("If the weight is an array, all its elements must be positive")          

        else:
            raise ValueError(f"A weight can only be a positive __array__, a positive float.")      

        return ArrayWeighting(array=weight, impl=impl, device=device, exponent=exponent)

    elif kwargs == {}:
        # TODO handle boolean case
        return ConstWeighting(const=1.0, impl=impl, device=device)

    elif kwargs == {'exponent': exponent}:
        return ConstWeighting(const=1.0, exponent=exponent, impl=impl, device=device)

    raise TypeError('got unknown keyword arguments {}'.format(kwargs))



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Unit tests for `discr_ops`."""

from __future__ import division

import numpy as np
import pytest

import odl
from odl.core.discr.discr_ops import _SUPPORTED_RESIZE_PAD_MODES
from odl.core.util.testutils import dtype_tol, noise_element, all_equal

from odl.core.util.dtype_utils import AVAILABLE_DTYPES, SCALAR_DTYPES, FLOAT_DTYPES, REAL_DTYPES
# --- pytest fixtures --- #


paddings = list(_SUPPORTED_RESIZE_PAD_MODES)
paddings.remove('constant')
paddings.extend([('constant', 0), ('constant', 1)])
padding_ids = [" pad_mode='{}'-{} ".format(*p)
               if isinstance(p, tuple)
               else " pad_mode='{}' ".format(p)
               for p in paddings]


@pytest.fixture(scope="module", ids=padding_ids, params=paddings)
def padding(request):
    if isinstance(request.param, tuple):
        pad_mode, pad_const = request.param
    else:
        pad_mode = request.param
        pad_const = 0

    return pad_mode, pad_const


# --- ResizingOperator tests --- #


def test_resizing_op_init(odl_impl_device_pairs, padding):
    # Test if the different init patterns run
    impl, device = odl_impl_device_pairs
    pad_mode, pad_const = padding

    space = odl.uniform_discr(
        [0, -1], [1, 1], (10, 5), impl=impl, device=device
        )
    res_space = odl.uniform_discr(
        [0, -3], [2, 3], (20, 15), impl=impl, device=device
    )

    odl.ResizingOperator(space, res_space)
    odl.ResizingOperator(space, ran_shp=(20, 15))
    odl.ResizingOperator(space, ran_shp=(20, 15), offset=(0, 5))
    odl.ResizingOperator(space, ran_shp=(20, 15), pad_mode=pad_mode)
    odl.ResizingOperator(space, ran_shp=(20, 15), pad_mode=pad_mode,
                         pad_const=pad_const)
    odl.ResizingOperator(space, ran_shp=(20, 15),
                         discr_kwargs={'nodes_on_bdry': True})


def test_resizing_op_raise(odl_impl_device_pairs):
    """Validate error checking in ResizingOperator."""
    # Domain not a uniformly discretized Lp
    with pytest.raises(TypeError):
        odl.ResizingOperator(odl.rn(5), ran_shp=(10,))

    grid = odl.RectGrid([0, 2, 3])
    part = odl.RectPartition(odl.IntervalProd(0, 3), grid)

    impl, device = odl_impl_device_pairs

    tspace = odl.rn(3, impl=impl, device=device)
    space = odl.DiscretizedSpace(part, tspace)
    with pytest.raises(ValueError):
        odl.ResizingOperator(space, ran_shp=(10,))

    # Different cell sides in domain and range
    space = odl.uniform_discr(0, 1, 10, impl=impl, device=device)
    res_space = odl.uniform_discr(0, 1, 15, impl=impl, device=device)
    with pytest.raises(ValueError):
        odl.ResizingOperator(space, res_space)

    # Non-integer multiple of cell sides used as shift (grid of the
    # resized space shifted)
    space = odl.uniform_discr(0, 1, 5, impl=impl, device=device)
    res_space = odl.uniform_discr(-0.5, 1.5, 10, impl=impl, device=device)
    with pytest.raises(ValueError):
        odl.ResizingOperator(space, res_space)

    # Need either range or ran_shp
    with pytest.raises(ValueError):
        odl.ResizingOperator(space)

    # Offset cannot be combined with range
    space = odl.uniform_discr([0, -1], [1, 1], (10, 5), impl=impl, device=device)
    res_space = odl.uniform_discr([0, -3], [2, 3], (20, 15), impl=impl, device=device)
    with pytest.raises(ValueError):
        odl.ResizingOperator(space, res_space, offset=(0, 0))

    # Bad pad_mode
    with pytest.raises(ValueError):
        odl.ResizingOperator(space, res_space, pad_mode='something')


def test_resizing_op_properties(odl_impl_device_pairs, padding):

    impl, device = odl_impl_device_pairs

    pad_mode, pad_const = padding

    for dtype in SCALAR_DTYPES:
        # Explicit range
        space = odl.uniform_discr([0, -1], [1, 1], (10, 5), dtype=dtype, impl=impl, device=device)
        res_space = odl.uniform_discr([0, -3], [2, 3], (20, 15), dtype=dtype, impl=impl, device=device)
        res_op = odl.ResizingOperator(space, res_space, pad_mode=pad_mode,
                                      pad_const=pad_const)

        assert res_op.domain == space
        assert res_op.range == res_space
        assert res_op.offset == (0, 5)
        assert res_op.pad_mode == pad_mode
        assert res_op.pad_const == pad_const
        if pad_mode == 'constant' and pad_const != 0:
            assert not res_op.is_linear
        else:
            assert res_op.is_linear

        # Implicit range via ran_shp and offset
        res_op = odl.ResizingOperator(space, ran_shp=(20, 15), offset=[0, 5],
                                      pad_mode=pad_mode, pad_const=pad_const)
        assert np.allclose(res_op.range.min_pt, res_space.min_pt)
        assert np.allclose(res_op.range.max_pt, res_space.max_pt)
        assert np.allclose(res_op.range.cell_sides, res_space.cell_sides)
        assert res_op.range.dtype == res_space.dtype
        assert res_op.offset == (0, 5)
        assert res_op.pad_mode == pad_mode
        assert res_op.pad_const == pad_const
        if pad_mode == 'constant' and pad_const != 0:
            assert not res_op.is_linear
        else:
            assert res_op.is_linear


def test_resizing_op_call(odl_impl_device_pairs):

    impl, device = odl_impl_device_pairs
    
    for dtype in AVAILABLE_DTYPES:
        # Minimal test since this operator only wraps resize_array
        space = odl.uniform_discr(
            [0, -1], [1, 1], (4, 5), dtype=dtype, impl=impl, device=device
        )
        res_space = odl.uniform_discr(
            [0, -0.6], [2, 0.2], (8, 2), dtype=dtype, impl=impl, device=device
        )
        res_op = odl.ResizingOperator(space, res_space)
        out = res_op(space.one())
        true_res = np.zeros((8, 2), dtype=dtype)
        true_res[:4, :] = 1
        assert all_equal(out, true_res)

        out = res_space.element()
        res_op(space.one(), out=out)
        assert all_equal(out, true_res)

        # Test also mapping to default impl for other 'impl'
        # if impl != 'numpy':
        #     space = odl.uniform_discr(
        #         [0, -1], [1, 1], (4, 5), dtype=dtype, impl=impl
        #     )
        #     res_space = odl.uniform_discr(
        #         [0, -0.6], [2, 0.2], (8, 2), dtype=dtype
        #     )
        #     res_op = odl.ResizingOperator(space, res_space)
        #     out = res_op(space.one())
        #     true_res = np.zeros((8, 2), dtype=dtype)
        #     true_res[:4, :] = 1
        #     assert all_equal(out, true_res)

        #     out = res_space.element()
        #     res_op(space.one(), out=out)
        #     assert all_equal(out, true_res)


def test_resizing_op_deriv(padding, odl_impl_device_pairs):

    impl, device = odl_impl_device_pairs

    pad_mode, pad_const = padding
    space = odl.uniform_discr(
        [0, -1], [1, 1], (4, 5), impl=impl, device=device
        )
    res_space = odl.uniform_discr(
        [0, -0.6], [2, 0.2], (8, 2), impl=impl, device=device
        )
    res_op = odl.ResizingOperator(space, res_space, pad_mode=pad_mode,
                                  pad_const=pad_const)
    res_op_deriv = res_op.derivative(space.one())

    if pad_mode == 'constant' and pad_const != 0:
        # Only non-trivial case is constant padding with const != 0
        assert res_op_deriv.pad_mode == 'constant'
        assert res_op_deriv.pad_const == 0.0
    else:
        assert res_op_deriv is res_op


def test_resizing_op_inverse(padding, odl_impl_device_pairs):

    impl, device = odl_impl_device_pairs
    pad_mode, pad_const = padding

    for dtype in SCALAR_DTYPES:

        if pad_mode == 'order1' and (
                np.issubdtype(dtype, np.unsignedinteger)
                or np.issubdtype(dtype, np.timedelta64()) ):
            # Extrapolating a trend might lead to negative values, which  
            # will raise an error for unsigned integers. For timedeltas, 
            # it would involve a multiplication of two times which was 
            # allowed by numpy 1 but is not allowed in numpy 2.
            continue

        space = odl.uniform_discr([0, -1], [1, 1], (4, 5), dtype=dtype,
                                  impl=impl, device=device)
        res_space = odl.uniform_discr([0, -1.4], [1.5, 1.4], (6, 7),
                                      dtype=dtype, impl=impl, device=device)
        res_op = odl.ResizingOperator(space, res_space, pad_mode=pad_mode,
                                      pad_const=pad_const)

        # Only left inverse if the operator extends in all axes
        x = noise_element(space)
        assert res_op.inverse(res_op(x)) == x


def test_resizing_op_adjoint(padding, odl_impl_device_pairs):

    impl, device = odl_impl_device_pairs
    pad_mode, pad_const = padding
    for dtype in FLOAT_DTYPES:
        space = odl.uniform_discr([0, -1], [1, 1], (4, 5), dtype=dtype,
                                  impl=impl, device=device)
        res_space = odl.uniform_discr([0, -1.4], [1.5, 1.4], (6, 7),
                                      dtype=dtype, impl=impl, device=device)
        res_op = odl.ResizingOperator(space, res_space, pad_mode=pad_mode,
                                      pad_const=pad_const)

        if pad_const != 0.0:
            with pytest.raises(NotImplementedError):
                res_op.adjoint
            return

        elem = noise_element(space)
        res_elem = noise_element(res_space)
        inner1 = res_op(elem).inner(res_elem)
        inner2 = elem.inner(res_op.adjoint(res_elem))
        assert inner1 == pytest.approx(
            inner2,
            abs = 1e-2 * space.size * dtype_tol(dtype) * elem.norm() * res_elem.norm())


def test_resizing_op_mixed_uni_nonuni(odl_impl_device_pairs):
    """Check if resizing along uniform axes in mixed discretizations works."""

    impl, device = odl_impl_device_pairs

    nonuni_part = odl.nonuniform_partition([0, 1, 4])
    uni_part = odl.uniform_partition(-1, 1, 4)
    part = uni_part.append(nonuni_part, uni_part, nonuni_part)
    tspace = odl.rn(part.shape, impl=impl, device=device)
    space = odl.DiscretizedSpace(part, tspace)

    # Keep non-uniform axes fixed
    res_op = odl.ResizingOperator(space, ran_shp=(6, 3, 6, 3))

    assert res_op.axes == (0, 2)
    assert res_op.offset == (1, 0, 1, 0)

    # Evaluation test with a simpler case
    part = uni_part.append(nonuni_part)
    tspace = odl.rn(part.shape, impl=impl, device=device)
    space = odl.DiscretizedSpace(part, tspace)
    res_op = odl.ResizingOperator(space, ran_shp=(6, 3))
    result = res_op(space.one())
    true_result = [[0, 0, 0],
                   [1, 1, 1],
                   [1, 1, 1],
                   [1, 1, 1],
                   [1, 1, 1],
                   [0, 0, 0]]
    assert all_equal(result, true_result)

    # Test adjoint
    elem = noise_element(space)
    res_elem = noise_element(res_op.range)
    inner1 = res_op(elem).inner(res_elem)
    inner2 = elem.inner(res_op.adjoint(res_elem))
    assert inner1 == pytest.approx(inner2)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Utility functions for space implementations."""

from __future__ import print_function, division, absolute_import
import numpy as np

from odl.core.util.npy_compat import AVOID_UNNECESSARY_COPY
from odl.core.array_API_support import lookup_array_backend

from odl.core.space.base_tensors import default_dtype

from odl.core.util.dtype_utils import is_available_dtype, is_complex_dtype, is_floating_dtype
from odl.core.space.entry_points import tensor_space_impl, tensor_space_impl_names

__all__ = ('vector', 'tensor_space', 'cn', 'rn')


def vector(array, dtype=None, impl='numpy', device = 'cpu'):
    """Create a vector from an array-like object.

    Parameters
    ----------
    array : `array-like`
        Array from which to create the vector. Scalars become
        one-dimensional vectors.
    dtype : optional
        Set the data type of the vector manually with this option.
        By default, the space type is inferred from the input data.
    impl : str, optional
        Impmlementation back-end for the space. See
        `odl.core.space.entry_points.tensor_space_impl_names` for available
        options.

    Returns
    -------
    vector : `Tensor`
        Vector created from the input array. Its concrete type depends
        on the provided arguments.

    Notes
    -----
    This is a convenience function and not intended for use in
    speed-critical algorithms.

    Examples
    --------
    Create one-dimensional vectors:

    >>> odl.vector([1, 2, 3])  # No automatic cast to float
    tensor_space(3, 'int64').element([1, 2, 3])
    >>> odl.vector([1, 2, 3], dtype=float)
    rn(3).element([ 1.,  2.,  3.])
    >>> odl.vector([1, 2 - 1j, 3])
    cn(3).element([ 1.+0.j,  2.-1.j,  3.+0.j])

    Non-scalar types are also supported:

    >>> odl.vector([True, True, False])
    tensor_space(3, 'bool').element([ True,  True, False])

    The function also supports multi-dimensional input:

    >>> odl.vector([[1, 2, 3],
    ...             [4, 5, 6]])
    tensor_space((2, 3), 'int64').element(
        [[1, 2, 3],
         [4, 5, 6]]
    )
    """
    backend = lookup_array_backend(impl)
    # Sanitize input
    # I don't understand was a ndim prepended to the array dimension
    arr = backend.array_constructor(array, copy=AVOID_UNNECESSARY_COPY, device = device)

    if arr.dtype is object:
        raise ValueError('invalid input data resulting in `dtype==object`')

    # Set dtype
    if dtype is not None:
        space_dtype = dtype
    else:
        space_dtype = arr.dtype

    space = tensor_space(arr.shape, dtype=space_dtype, impl=impl, device=device)
    return space.element(arr)


def tensor_space(shape, dtype='float64', impl='numpy', device = 'cpu', **kwargs):
    """Return a tensor space with arbitrary scalar data type.

    Parameters
    ----------
    shape : positive int or sequence of positive ints
        Number of entries per axis for elements in this space. A
        single integer results in a space with 1 axis.
    dtype (str) : optional
        Data type of each element. Defaults to float64
    impl : str, optional
        Impmlementation back-end for the space. See
        `odl.core.space.entry_points.tensor_space_impl_names` for available
        options.
    kwargs :
        Extra keyword arguments passed to the space constructor.

    Returns
    -------
    space : `TensorSpace`

    Examples
    --------
    Space of 3-tuples with ``uint64`` entries (although not strictly a
    vector space):

    >>> odl.tensor_space(3, dtype='uint64')
    tensor_space(3, 'uint64')

    2x3 tensors with same data type:

    >>> odl.tensor_space((2, 3), dtype='uint64')
    tensor_space((2, 3), 'uint64')

    The default data type is ``'float64'``. How that is represented as a dtype-object
    depends on the backend.

    >>> ts = odl.tensor_space((2, 3))
    >>> ts
    rn((2, 3))
    >>> ts.dtype
    dtype('float64')

    See Also
    --------
    rn, cn : Constructors for real and complex spaces
    """
    # Check the dtype argument
    is_available_dtype(dtype)
    # Check the impl argument
    assert (
        impl in tensor_space_impl_names()
    ), f"The only supported impls are {tensor_space_impl_names()}, but {impl} was provided"

    # Use args by keyword since the constructor may take other arguments
    # by position
    return tensor_space_impl(impl)(shape=shape, dtype=dtype, device=device, **kwargs)


def cn(shape, dtype='complex128', impl='numpy', device='cpu', **kwargs):
    """Return a space of complex tensors.

    Parameters
    ----------
    shape : positive int or sequence of positive ints
        Number of entries per axis for elements in this space. A
        single integer results in a space with 1 axis.
    dtype (str) : optional
        Data type of each element. Must be provided as a string or Python complex type.
        Defaults to complex128
    impl (str) : str, optional
        Impmlementation back-end for the space. See
        `odl.core.space.entry_points.tensor_space_impl_names` for available
        options.
    kwargs :
        Extra keyword arguments passed to the space constructor.

    Returns
    -------
    cn : `TensorSpace`

    Examples
    --------
    Space of complex 3-tuples with ``complex64`` entries:

    >>> odl.cn(3, dtype='complex64')
    cn(3, 'complex64')

    Complex 2x3 tensors with ``complex64`` entries:

    >>> odl.cn((2, 3), dtype='complex64')
    cn((2, 3), 'complex64')

    The default data type depends on the implementation. For
    ``impl='numpy'``, it is ``'complex128'``:

    >>> space = odl.cn((2, 3))
    >>> space
    cn((2, 3))
    >>> space.dtype
    dtype('complex128')

    See Also
    --------
    tensor_space : Space of tensors with arbitrary scalar data type.
    rn : Real tensor space.
    """
    is_complex_dtype(dtype)
    return tensor_space(shape, dtype=dtype, impl=impl, device=device, **kwargs)


def rn(shape, dtype=None, impl='numpy', device ='cpu', **kwargs):
    """Return a space of real tensors.

    Parameters
    ----------
    shape : positive int or sequence of positive ints
        Number of entries per axis for elements in this space. A
        single integer results in a space with 1 axis.
    dtype (str) : optional
        Data type of each element. See REAL_DTYPES in 
        `odl.core.util.utility.py` for available options. Defaults to float64
    impl (str) : str, optional
        Impmlementation back-end for the space. See the constant
        TENSOR_SPACE_IMPLS for available backends
        options.
    kwargs :
        Extra keyword arguments passed to the space constructor.

    Returns
    -------
    real_space : `TensorSpace`

    Examples
    --------
    Space of real 3-tuples with ``float32`` entries:

    >>> odl.rn(3, dtype='float32')
    rn(3, 'float32')

    Real 2x3 tensors with ``float32`` entries:

    >>> odl.rn((2, 3), dtype='float32')
    rn((2, 3), 'float32')

    The default data type is float64

    >>> ts = odl.rn((2, 3))
    >>> ts
    rn((2, 3))
    >>> ts.dtype
    dtype('float64')

    See Also
    --------
    tensor_space : Space of tensors with arbitrary scalar data type.
    cn : Complex tensor space.
    """
    if dtype is None:
        dtype = default_dtype(lookup_array_backend(str(impl).lower()))
    is_floating_dtype(dtype)
    return tensor_space(shape, dtype=dtype, impl=impl, device=device, **kwargs)



if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Large-scale tests for the Fourier transform.

This test suite is intended for performance monitoring, to capture
speed regressions.
"""

from __future__ import division

import numpy as np
import pytest

import odl
from odl.core.util.testutils import simple_fixture, skip_if_no_pyfftw

# --- pytest fixtures --- #


# Global pytest mark
pytestmark = pytest.mark.suite('largescale')


impl = simple_fixture(
    'impl',
    [pytest.param('numpy'),
     pytest.param('pyfftw', marks=skip_if_no_pyfftw)]
)

domain = simple_fixture(
    name='domain',
    params=[odl.uniform_discr(-2, 2, 10 ** 5),
            odl.uniform_discr([-2, -2, -2], [2, 2, 2], [200, 200, 200]),
            odl.uniform_discr(-2, 2, 10 ** 5, dtype=complex),
            odl.uniform_discr([-2, -2, -2], [2, 2, 2], [200, 200, 200],
                              dtype=complex)])


# --- FourierTransform tests --- #


def test_dft_forward(domain, impl):

    halfcomplex = np.issubdtype(domain.dtype, np.floating)
    dft = odl.trafos.DiscreteFourierTransform(domain, halfcomplex=halfcomplex)
    one = domain.one()
    out = dft.range.element()

    dft(one, out=out)
    assert out[(0,) * out.ndim] == pytest.approx(domain.size)


def test_fourier_trafo_forward_complex(domain, impl):

    if domain.field == odl.RealNumbers():
        return

    ft = odl.trafos.FourierTransform(domain, impl=impl)

    def charfun_ball(x):
        sum_sq = sum(xi ** 2 for xi in x)
        return np.where(sum_sq < 1, 1, 0)

    def charfun_freq_ball(x):
        rad = np.max(ft.range.domain.extent / 4)
        sum_sq = sum(xi ** 2 for xi in x)
        return np.where(sum_sq < rad ** 2, 1, 0)

    ball_dom = ft.domain.element(charfun_ball)
    ball_ran = ft.range.element(charfun_freq_ball)

    ball_dom_ft = ft(ball_dom)
    ball_ran_ift = ft.adjoint(ball_ran)
    assert (
        ball_dom.inner(ball_ran_ift)
        == pytest.approx(ball_ran.inner(ball_dom_ft), rel=0.1)
    )


if __name__ == '__main__':
    odl.core.util.test_file(__file__, ['-S', 'largescale'])



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Alternating dual (AD) update algorithm studied by McGaffin and Fessler.

The alternating dual upgrade method solves structured convex optimization
problems by successively updating dual variables which are associated with
each of the components.
"""

from __future__ import print_function, division, absolute_import

import numpy as np

__all__ = ('adupdates',)


def adupdates(x, g, L, stepsize, inner_stepsizes, niter, random=False,
              callback=None, callback_loop='outer'):
    r"""Alternating Dual updates method.

    The Alternating Dual (AD) updates method of McGaffin and Fessler `[MF2015]
    <http://ieeexplore.ieee.org/document/7271047/>`_ is designed to solve an
    optimization problem of the form ::

        min_x [ sum_i g_i(L_i x) ]

    where ``g_i`` are proper, convex and lower semicontinuous functions and
    ``L_i`` are linear `Operator` s.

    Parameters
    ----------
    g : sequence of `Functional` s
        All functions need to provide a `Functional.convex_conj` with a
        `Functional.proximal` factory.
    L : sequence of `Operator` s
        Length of ``L`` must equal the length of ``g``.
    x : `LinearSpaceElement`
        Initial point, updated in-place.
    stepsize : positive float
        The stepsize for the outer (proximal point) iteration. The theory
        guarantees convergence for any positive real number, but the
        performance might depend on the choice of a good stepsize.
    inner_stepsizes : sequence of stepsizes
        Parameters determining the stepsizes for the inner iterations. Must be
        matched with the norms of ``L``, and convergence is guaranteed if the
        ``inner_stepsizes`` are small enough. See the Notes section for
        details.
    niter : int
        Number of (outer) iterations.
    random : bool, optional
        If `True`, the order of the dual upgdates is chosen randomly,
        otherwise the order provided by the lists ``g``, ``L`` and
        ``inner_stepsizes`` is used.
    callback : callable, optional
        Function called with the current iterate after each iteration.
    callback_loop : {'inner', 'outer'}, optional
       If 'inner', the ``callback`` function is called after each inner
       iteration, i.e., after each dual update. If 'outer', the ``callback``
       function is called after each outer iteration, i.e., after each primal
       update.

    Notes
    -----
    The algorithm as implemented here is described in the article [MF2015],
    where it is applied to a tomography problem. It solves the problem

    .. math::
        \min_x \sum_{i=1}^m g_i(L_i x),

    where :math:`g_i` are proper, convex and lower semicontinuous functions
    and :math:`L_i` are linear, continuous operators for
    :math:`i = 1, \ldots, m`. In an outer iteration, the solution is found
    iteratively by an iteration

    .. math::
        x_{n+1} = \mathrm{arg\,min}_x \sum_{i=1}^m g_i(L_i x)
            + \frac{\mu}{2} \|x - x_n\|^2

    with some ``stepsize`` parameter :math:`\mu > 0` according to the proximal
    point algorithm. In the inner iteration, dual variables are introduced for
    each of the components of the sum. The Lagrangian of the problem is given
    by

    .. math::
        S_n(x; v_1, \ldots, v_m) = \sum_{i=1}^m (\langle v_i, L_i x \rangle
            - g_i^*(v_i)) + \frac{\mu}{2} \|x - x_n\|^2.

    Given the dual variables, the new primal variable :math:`x_{n+1}`
    can be calculated by directly minimizing :math:`S_n` with respect to
    :math:`x`. This corresponds to the formula

    .. math::
         x_{n+1} = x_n - \frac{1}{\mu} \sum_{i=1}^m L_i^* v_i.

    The dual updates are executed according to the following rule:

    .. math::
        v_i^+ = \mathrm{Prox}^{\mu M_i^{-1}}_{g_i^*}
            (v_i + \mu M_i^{-1} L_i x_{n+1}),

    where :math:`x_{n+1}` is given by the formula above and  :math:`M_i` is a
    diagonal matrix with positive diagonal entries such that
    :math:`M_i - L_i L_i^*` is positive semidefinite. The variable
    ``inner_stepsizes`` is chosen as a stepsize to the `Functional.proximal` to
    the `Functional.convex_conj` of each of the ``g`` s after multiplying with
    ``stepsize``. The ``inner_stepsizes`` contain the elements of
    :math:`M_i^{-1}` in one of the following ways:

    * Setting ``inner_stepsizes[i]`` a positive float :math:`\gamma`
      corresponds to the choice :math:`M_i^{-1} = \gamma \mathrm{Id}`.
    * Assume that ``g_i`` is a `SeparableSum`, then setting
      ``inner_stepsizes[i]`` a list :math:`(\gamma_1, \ldots, \gamma_k)` of
      positive floats corresponds to the choice of a block-diagonal matrix
      :math:`M_i^{-1}`, where each block corresponds to one of the space
      components and equals :math:`\gamma_i \mathrm{Id}`.
    * Assume that ``g_i`` is an `L1Norm` or an `L2NormSquared`, then setting
      ``inner_stepsizes[i]`` a ``g_i.domain.element`` :math:`z` corresponds to
      the choice :math:`M_i^{-1} = \mathrm{diag}(z)`.

    References
    ----------
    [MF2015] McGaffin, M G, and Fessler, J A. *Alternating dual updates
    algorithm for X-ray CT reconstruction on the GPU*. IEEE Transactions
    on Computational Imaging, 1.3 (2015), pp 186--199.
    """
    # Check the lenghts of the lists (= number of dual variables)
    length = len(g)
    if len(L) != length:
        raise ValueError('`len(L)` should equal `len(g)`, but {} != {}'
                         ''.format(len(L), length))

    if len(inner_stepsizes) != length:
        raise ValueError('len(`inner_stepsizes`) should equal `len(g)`, '
                         ' but {} != {}'.format(len(inner_stepsizes), length))

    # Check if operators have a common domain
    # (the space of the primal variable):
    domain = L[0].domain
    if any(opi.domain != domain for opi in L):
        raise ValueError('domains of `L` are not all equal')

    # Check if range of the operators equals domain of the functionals
    ranges = [opi.range for opi in L]
    if any(L[i].range != g[i].domain for i in range(length)):
        raise ValueError('L[i].range` should equal `g.domain`')

    # Normalize string
    callback_loop, callback_loop_in = str(callback_loop).lower(), callback_loop
    if callback_loop not in ('inner', 'outer'):
        raise ValueError('`callback_loop` {!r} not understood'
                         ''.format(callback_loop_in))

    # Initialization of the dual variables
    duals = [space.zero() for space in ranges]

    # Reusable elements in the ranges, one per type of space
    unique_ranges = set(ranges)
    tmp_rans = {ran: ran.element() for ran in unique_ranges}

    # Prepare the proximal operators. Since the stepsize does not vary over
    # the iterations, we always use the same proximal operator.
    proxs = [func.convex_conj.proximal(stepsize * inner_ss
                                       if np.isscalar(inner_ss)
                                       else stepsize * np.asarray(inner_ss))
             for (func, inner_ss) in zip(g, inner_stepsizes)]

    # Iteratively find a solution
    for _ in range(niter):
        # Update x = x - 1/stepsize * sum([ops[i].adjoint(duals[i])
        # for i in range(length)])
        for i in range(length):
            x -= (1.0 / stepsize) * L[i].adjoint(duals[i])

        if random:
            rng = np.random.permutation(range(length))
        else:
            rng = range(length)

        for j in rng:
            step = (stepsize * inner_stepsizes[j]
                    if np.isscalar(inner_stepsizes[j])
                    else stepsize * np.asarray(inner_stepsizes[j]))
            arg = duals[j] + step * L[j](x)
            tmp_ran = tmp_rans[L[j].range]
            proxs[j](arg, out=tmp_ran)
            x -= 1.0 / stepsize * L[j].adjoint(tmp_ran - duals[j])
            duals[j].assign(tmp_ran)

            if callback is not None and callback_loop == 'inner':
                callback(x)
        if callback is not None and callback_loop == 'outer':
            callback(x)


def adupdates_simple(x, g, L, stepsize, inner_stepsizes, niter,
                     random=False):
    """Non-optimized version of ``adupdates``.
    This function is intended for debugging. It makes a lot of copies and
    performs no error checking.
    """
    # Initializations
    length = len(g)
    ranges = [Li.range for Li in L]
    duals = [space.zero() for space in ranges]

    # Iteratively find a solution
    for _ in range(niter):
        # Update x = x - 1/stepsize * sum([ops[i].adjoint(duals[i])
        # for i in range(length)])
        for i in range(length):
            x -= (1.0 / stepsize) * L[i].adjoint(duals[i])

        rng = np.random.permutation(range(length)) if random else range(length)

        for j in rng:
            dual_tmp = ranges[j].element()
            dual_tmp = (g[j].convex_conj.proximal
                        (stepsize * inner_stepsizes[j]
                         if np.isscalar(inner_stepsizes[j])
                         else stepsize * np.asarray(inner_stepsizes[j]))
                        (duals[j] + stepsize * inner_stepsizes[j] * L[j](x)
                         if np.isscalar(inner_stepsizes[j])
                         else duals[j] + stepsize *
                         np.asarray(inner_stepsizes[j]) * L[j](x)))
            x -= 1.0 / stepsize * L[j].adjoint(dual_tmp - duals[j])
            duals[j].assign(dual_tmp)



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Weightings for finite-dimensional spaces."""

from __future__ import print_function, division, absolute_import
from builtins import object
import math
import numpy as np

from odl.core.util import array_str, signature_string, indent, is_real_dtype
from odl.core.array_API_support.utils import get_array_and_backend, lookup_array_backend
from odl.core.array_API_support.comparisons import odl_all_equal

__all__ = ('MatrixWeighting', 'ArrayWeighting', 'ConstWeighting',
           'CustomInner', 'CustomNorm', 'CustomDist')


class Weighting(object):

    """Abstract base class for weighting of finite-dimensional spaces.

    This class and its subclasses serve as a simple means to evaluate
    and compare weighted inner products, norms and metrics semantically
    rather than by identity on a pure function level.

    The functions are implemented similarly to `Operator`,
    but without extra type checks of input parameters - this is done in
    the callers of the `LinearSpace` instance where these
    functions are being used.
    """

    def __init__(self, impl, device, exponent=2.0):
        """Initialize a new instance.

        Parameters
        ----------
        impl : string
            Specifier for the implementation backend
        device :
            device identifier, compatible with the backend associated with `impl`
        exponent : positive float, optional
            Exponent of the norm. For values other than 2.0, the inner
            product is not defined.
        """
        self.__impl = str(impl).lower()
        self.__exponent = float(exponent)
        self.__device = device
        if self.exponent <= 0:
            raise ValueError('only positive exponents or inf supported, '
                             'got {}'.format(exponent))

    @property
    def impl(self):
        """Implementation backend of this weighting."""
        return self.__impl

    @property
    def device(self):
        """Backend-specific device identifier. Arrays this weighting should measure
        must be stored on that device."""
        return self.__device

    def to_device(self, device):
        """Return a version of the same weighting, but with any internal arrays stored
        on a different device."""
        raise NotImplementedError("Abstract method")

    def to_impl(self, impl):
        """Return a version of the same weighting, but with any internal arrays stored
        with a different array backend (e.g. `'pytorch'` instead of `'numpy'`)."""
        raise NotImplementedError("Abstract method")

    @property
    def shape(self):
        """A tuple of numbers, denoting the shape that arrays need to have to be
        used with this weighting. An empty shape means any shape of array is supported."""
        raise NotImplementedError("Abstract method")

    @property
    def exponent(self):
        """Exponent of this weighting."""
        return self.__exponent

    def __eq__(self, other):
        """Return ``self == other``.

        Returns
        -------
        equal : bool
            ``True`` if ``other`` is a the same weighting, ``False``
            otherwise.

        Notes
        -----
        This operation must be computationally cheap, i.e. no large
        arrays may be compared entry-wise. That is the task of the
        `equiv` method.
        """
        return (isinstance(other, Weighting) and
                self.impl == other.impl and
                self.exponent == other.exponent)

    def __hash__(self):
        """Return ``hash(self)``."""
        return hash((type(self), self.impl, self.exponent))

    def equiv(self, other):
        """Test if ``other`` is an equivalent weighting.

        Should be overridden, default tests for equality.

        Returns
        -------
        equivalent : bool
            ``True`` if ``other`` is a `Weighting` instance which
            yields the same result as this inner product for any
            input, ``False`` otherwise.
        """
        return self == other

    def inner(self, x1, x2):
        """Return the inner product of two elements.

        Parameters
        ----------
        x1, x2 : ArrayLike
            Elements whose inner product is calculated.

        Returns
        -------
        inner : float or complex
            The inner product of the two provided elements.
        """
        raise NotImplementedError

    def norm(self, x):
        """Calculate the norm of an element.

        This is the standard implementation using `inner`.
        Subclasses should override it for optimization purposes.

        Parameters
        ----------
        x1 : ArrayLike
            Element whose norm is calculated.

        Returns
        -------
        norm : float
            The norm of the element.
        """
        return float(math.sqrt(self.inner(x, x).real))

    def dist(self, x1, x2):
        """Calculate the distance between two elements.

        This is the standard implementation using `norm`.
        Subclasses should override it for optimization purposes.

        Parameters
        ----------
        x1, x2 : ArrayLike
            Elements whose mutual distance is calculated.

        Returns
        -------
        dist : float
            The distance between the elements.
        """
        return self.norm(x1 - x2)


class MatrixWeighting(Weighting):

    """Weighting of a space by a matrix.

    The exact definition of the weighted inner product, norm and
    distance functions depend on the concrete space.

    The matrix must be Hermitian and posivive definite, otherwise it
    does not define an inner product or norm, respectively. This is not
    checked during initialization.
    """

    def __init__(self, matrix, impl, device, exponent=2.0, **kwargs):
        """Initialize a new instance.

        Parameters
        ----------
        matrix :  `scipy.sparse.spmatrix` or 2-dim. `array-like`
            Square weighting matrix of the inner product
        impl : string
            Specifier for the implementation backend
        device :
            device identifier, compatible with the backend associated with `impl`
        exponent : positive float, optional
            Exponent of the norm. For values other than 2.0, the inner
            product is not defined.
            If ``matrix`` is a sparse matrix, only 1.0, 2.0 and ``inf``
            are allowed.
        precomp_mat_pow : bool, optional
            If ``True``, precompute the matrix power ``W ** (1/p)``
            during initialization. This has no effect if ``exponent``
            is 1.0, 2.0 or ``inf``.

            Default: ``False``

        cache_mat_pow : bool, optional
            If ``True``, cache the matrix power ``W ** (1/p)``. This can
            happen either during initialization or in the first call to
            ``norm`` or ``dist``, resp. This has no effect if
            ``exponent`` is 1.0, 2.0 or ``inf``.

            Default: ``True``

        cache_mat_decomp : bool, optional
            If ``True``, cache the eigenbasis decomposition of the
            matrix. This can happen either during initialization or in
            the first call to ``norm`` or ``dist``, resp. This has no
            effect if ``exponent`` is 1.0, 2.0 or ``inf``.

            Default: ``False``

        Notes
        -----
        The matrix power ``W ** (1/p)`` is computed by eigenbasis
        decomposition::

            eigval, eigvec = scipy.linalg.eigh(matrix)
            mat_pow = (eigval ** p * eigvec).dot(eigvec.conj().T)

        Depending on the matrix size, this can be rather expensive.
        """
        # Lazy import to improve `import odl` time
        import scipy.sparse

        # TODO: fix dead link `scipy.sparse.spmatrix`
        precomp_mat_pow = kwargs.pop('precomp_mat_pow', False)
        self._cache_mat_pow = bool(kwargs.pop('cache_mat_pow', True))
        self._cache_mat_decomp = bool(kwargs.pop('cache_mat_decomp', False))
        super(MatrixWeighting, self).__init__(impl=impl, device=device, exponent=exponent)

        # Check and set matrix
        if scipy.sparse.isspmatrix(matrix):
            self._matrix = matrix
        else:
            self._matrix = np.asarray(matrix)
            if self._matrix.dtype == object:
                raise ValueError('invalid matrix {}'.format(matrix))
            elif self._matrix.ndim != 2:
                raise ValueError('matrix {} is {}-dimensional instead of '
                                 '2-dimensional'
                                 ''.format(matrix, self._matrix.ndim))

        if self._matrix.shape[0] != self._matrix.shape[1]:
            raise ValueError('matrix has shape {}, expected a square matrix'
                             ''.format(self._matrix.shape))

        if (scipy.sparse.isspmatrix(self.matrix) and
                self.exponent not in (1.0, 2.0, float('inf'))):
            raise NotImplementedError('sparse matrices only supported for '
                                      'exponent 1.0, 2.0 or `inf`')

        # Compute the power and decomposition if desired
        self._eigval = self._eigvec = None
        if self.exponent in (1.0, float('inf')):
            self._mat_pow = self.matrix
        elif precomp_mat_pow and self.exponent != 2.0:
            eigval, eigvec = self.matrix_decomp()
            if self._cache_mat_decomp:
                self._eigval, self._eigvec = eigval, eigvec
                eigval_pow = eigval ** (1.0 / self.exponent)
            else:
                eigval_pow = eigval
                eigval_pow **= 1.0 / self.exponent
            self._mat_pow = (eigval_pow * eigvec).dot(eigvec.conj().T)
        else:
            self._mat_pow = None

    @property
    def matrix(self):
        """Weighting matrix of this inner product."""
        return self._matrix

    def is_valid(self):
        """Test if the matrix is positive definite Hermitian.

        If the matrix decomposition is available, this test checks
        if all eigenvalues are positive.
        Otherwise, the test tries to calculate a Cholesky decomposition,
        which can be very time-consuming for large matrices. Sparse
        matrices are not supported.
        """
        # Lazy import to improve `import odl` time
        import scipy.sparse

        if scipy.sparse.isspmatrix(self.matrix):
            raise NotImplementedError('validation not supported for sparse '
                                      'matrices')
        elif self._eigval is not None:
            return np.all(np.greater(self._eigval, 0))
        else:
            try:
                np.linalg.cholesky(self.matrix)
                return np.array_equal(self.matrix, self.matrix.conj().T)
            except np.linalg.LinAlgError:
                return False

    def matrix_decomp(self, cache=None):
        """Compute a Hermitian eigenbasis decomposition of the matrix.

        Parameters
        ----------
        cache : bool or None, optional
            If ``True``, store the decomposition internally. For None,
            the ``cache_mat_decomp`` from class initialization is used.

        Returns
        -------
        eigval : `numpy.ndarray`
            One-dimensional array of eigenvalues. Its length is equal
            to the number of matrix rows.
        eigvec : `numpy.ndarray`
            Two-dimensional array of eigenvectors. It has the same shape
            as the decomposed matrix.

        See Also
        --------
        scipy.linalg.decomp.eigh :
            Implementation of the decomposition. Standard parameters
            are used here.

        Raises
        ------
        NotImplementedError
            if the matrix is sparse (not supported by scipy 0.17)
        """
        # Lazy import to improve `import odl` time
        import scipy.linalg
        import scipy.sparse

        # TODO: fix dead link `scipy.linalg.decomp.eigh`
        if scipy.sparse.isspmatrix(self.matrix):
            raise NotImplementedError('sparse matrix not supported')

        if cache is None:
            cache = self._cache_mat_decomp

        if self._eigval is None or self._eigvec is None:
            eigval, eigvec = scipy.linalg.eigh(self.matrix)
            if cache:
                self._eigval = eigval
                self._eigvec = eigvec
        else:
            eigval, eigvec = self._eigval, self._eigvec

        return eigval, eigvec

    def __eq__(self, other):
        """Return ``self == other``.

        Returns
        -------
        equals : bool
            ``True`` if other is a `MatrixWeighting` instance
            with **identical** matrix, ``False`` otherwise.

        See Also
        --------
        equiv : test for equivalent inner products
        """
        if other is self:
            return True

        return (super(MatrixWeighting, self).__eq__(other) and
                self.matrix is getattr(other, 'matrix', None))

    def __hash__(self):
        """Return ``hash(self)``."""
        # TODO: Better hash for matrix?
        return hash((super(MatrixWeighting, self).__hash__(),
                     self.matrix.tobytes()))

    def equiv(self, other):
        """Test if other is an equivalent weighting.

        Returns
        -------
        equivalent : bool
            ``True`` if ``other`` is a `Weighting` instance with the same
            `Weighting.impl`, which yields the same result as this
            weighting for any input, ``False`` otherwise. This is checked
            by entry-wise comparison of matrices/arrays/constants.
        """
        # Lazy import to improve `import odl` time
        import scipy.sparse

        # Optimization for equality
        if self == other:
            return True

        elif self.exponent != getattr(other, 'exponent', -1):
            return False

        elif isinstance(other, MatrixWeighting):
            if self.matrix.shape != other.matrix.shape:
                return False

            if scipy.sparse.isspmatrix(self.matrix):
                if other.matrix_issparse:
                    # Optimization for different number of nonzero elements
                    if self.matrix.nnz != other.matrix.nnz:
                        return False
                    else:
                        # Most efficient out-of-the-box comparison
                        return (self.matrix != other.matrix).nnz == 0
                else:  # Worst case: compare against dense matrix
                    return np.array_equal(self.matrix.todense(), other.matrix)

            else:  # matrix of `self` is dense
                if other.matrix_issparse:
                    return np.array_equal(self.matrix, other.matrix.todense())
                else:
                    return np.array_equal(self.matrix, other.matrix)

        elif isinstance(other, ArrayWeighting):
            if scipy.sparse.isspmatrix(self.matrix):
                return (np.array_equiv(self.matrix.diagonal(),
                                       other.array) and
                        np.array_equal(self.matrix.asformat('dia').offsets,
                                       np.array([0])))
            else:
                return np.array_equal(
                    self.matrix, other.array * np.eye(self.matrix.shape[0]))

        elif isinstance(other, ConstWeighting):
            if scipy.sparse.isspmatrix(self.matrix):
                return (np.array_equiv(self.matrix.diagonal(), other.const) and
                        np.array_equal(self.matrix.asformat('dia').offsets,
                                       np.array([0])))
            else:
                return np.array_equal(
                    self.matrix, other.const * np.eye(self.matrix.shape[0]))
        else:
            return False

    @property
    def repr_part(self):
        """Return a string usable in a space's ``__repr__`` method."""
        # Lazy import to improve `import odl` time
        import scipy.sparse

        if scipy.sparse.isspmatrix(self.matrix):
            optargs = [('matrix', str(self.matrix), '')]
        else:
            optargs = [('matrix', array_str(self.matrix, nprint=10), '')]

        optargs.append(('exponent', self.exponent, 2.0))
        return signature_string([], optargs, mod=[[], ['!s', ':.4']])

    def __repr__(self):
        """Return ``repr(self)``."""
        if self.matrix_issparse:
            posargs = ['<{} sparse matrix, format {}, {} nonzero entries>'
                       ''.format(self.matrix.shape, self.matrix.format,
                                 self.matrix.nnz)]
        else:
            posargs = [array_str(self.matrix, nprint=10)]

        optargs = [('exponent', self.exponent, 2.0)]
        inner_str = signature_string(posargs, optargs, sep=',\n',
                                     mod=['!s', ''])
        return '{}(\n{}\n)'.format(self.__class__.__name__, indent(inner_str))

    def __str__(self):
        """Return ``str(self)``."""
        return repr(self)


def _pnorm_diagweight(x, p, w):
    """Diagonally weighted p-norm implementation."""
    x, array_backend = get_array_and_backend(x)
    ns = array_backend.array_namespace
    xp = ns.abs(x.data)
    if p == float('inf'):
        xp *= w
        return ns.max(xp)
    else:
        # Believe it or not, Pytorch and Numpy implement power in a *different* way
        try:
            xp = ns.power(xp, p, out=xp)
        except AttributeError:
            xp = ns.pow(xp, p, out=xp)
        xp *= w
        return ns.sum(xp) ** (1 / p)

def _norm_default(x):
    """Default Euclidean norm implementation."""
    x, array_backend = get_array_and_backend(x)
    ns = array_backend.array_namespace
    return ns.linalg.vector_norm(x.data)

def _pnorm_default(x, p):
    """Default p-norm implementation."""
    x, array_backend = get_array_and_backend(x)
    ns = array_backend.array_namespace
    return ns.linalg.vector_norm(x.data, ord=p)

def _inner_default(x1, x2):
    """Default Euclidean inner product implementation."""
    x1, array_backend_1 = get_array_and_backend(x1)
    x2, array_backend_2 = get_array_and_backend(x2)
    assert  array_backend_1 == array_backend_2, f"{array_backend_1=} and {array_backend_2=} do not match"
    ns = array_backend_1.array_namespace
    if is_real_dtype(x2.dtype):
        return ns.vecdot(x1.ravel(), x2.ravel())
    else:
        # `vecdot` has the complex conjugate on the left argument,
        # whereas ODL convention is that the inner product should
        # be linear in the left argument (conjugate in the right).
        return ns.vecdot(x2.ravel(), x1.ravel())


# TODO: implement intermediate weighting schemes with arrays that are
# broadcast, i.e. between scalar and full-blown in dimensionality?

class ArrayWeighting(Weighting):

    """Weighting of a space by an array.

    The exact definition of the weighted inner product, norm and
    distance functions depend on the concrete space.

    The array may only have positive entries, otherwise it does not
    define an inner product or norm, respectively. This is not checked
    during initialization.
    """

    def __init__(self, array, impl, device, exponent=2.0):
        """Initialize a new instance.

        Parameters
        ----------
        array : `array-like`
            Weighting array of inner product, norm and distance.
            Native `Tensor` instances are stored as-is without copying.
            Do not pass an ODL-space-element here. If you want to use such
            an element, use its contained `data` instead.
        impl : string
            Specifier for the implementation backend.
        device :
            device identifier, compatible with the backend associated with `impl`
        exponent : positive float, optional
            Exponent of the norm. For values other than 2.0, the inner
            product is not defined.
        """
        super(ArrayWeighting, self).__init__(impl=impl, device=device, exponent=exponent)

        # We apply array duck-typing to allow all kinds of Numpy-array-like
        # data structures without change
        array_attrs = ('shape', 'dtype', 'itemsize')
        if (all(hasattr(array, attr) for attr in array_attrs)):
            self.__array = array
        # TODO add a check that the array is compatible with the `impl`, and if not either
        # convert it or raise an error. This should be done using Python Array API features.
        else:
            raise TypeError('`array` {!r} does not look like a valid array'
                            ''.format(array))

    @property
    def array(self):
        """Weighting array of this instance."""
        return self.__array

    @property
    def weight(self):
        """Weighting array of this instance."""
        return self.array

    @property
    def shape(self):
        """Arrays measured by this weighting must have the same shape as the
        weighting array itself."""
        return self.array.shape

    def to_device(self, device):
        _, backend = get_array_and_backend(self.array)
        return ArrayWeighting(array = backend.to_device(self.array, device=device), impl=self.impl, device=device, exponent=self.exponent)

    def to_impl(self, impl):
        new_backend = lookup_array_backend(impl)
        new_array = new_backend.array_namespace.from_dlpack(self.array)

        # TODO the following is likely to fail in case e.g. torch-cuda is sent to 'numpy'.
        # It is required to first use `to_device('cpu')`, then `to_impl`.
        # It would be useful to add a device argument that allows changing backend and device in
        # one step. This is currently hampered by missing `device` argument to `from_dlpack` in Torch.
        assert(new_array.device == self.device)

        return ArrayWeighting(array=new_array, impl=impl, device=self.device, exponent=self.exponent)

    def is_valid(self):
        """Return True if the array is a valid weight, i.e. positive."""
        return np.all(np.greater(self.array, 0))

    def __eq__(self, other):
        """Return ``self == other``.

        Returns
        -------
        equals : bool
            ``True`` if ``other`` is an `ArrayWeighting` instance with
            **identical** array, False otherwise.

        See Also
        --------
        equiv : test for equivalent inner products
        """
        if other is self:
            return True

        return (super(ArrayWeighting, self).__eq__(other) and
                odl_all_equal(self.array, other.array))

    def __hash__(self):
        """Return ``hash(self)``."""
        # TODO: Better hash for array?
        return hash((super(ArrayWeighting, self).__hash__(),
                     self.array.tobytes()))

    def equiv(self, other):
        """Return True if other is an equivalent weighting.

        Returns
        -------
        equivalent : bool
            ``True`` if ``other`` is a `Weighting` instance with the same
            `Weighting.impl`, which yields the same result as this
            weighting for any input, ``False`` otherwise. This is checked
            by entry-wise comparison of arrays/constants.
        """
        # Optimization for equality
        if (not isinstance(other, Weighting) or
              self.exponent != other.exponent):
            return False
        elif isinstance(other, MatrixWeighting):
            return other.equiv(self)
        elif isinstance(other, ConstWeighting):
            # return np.array_equiv(self.array, other.const)
            return odl_all_equal(self.array, other.const)
        else:
            # return np.array_equal(self.array, other.array)
            return odl_all_equal(self.array, other.array)

    @property
    def repr_part(self):
        """String usable in a space's ``__repr__`` method."""
        optargs = [('weighting', array_str(self.array, nprint=10), ''),
                   ('exponent', self.exponent, 2.0)]
        return signature_string([], optargs, sep=',\n',
                                mod=[[], ['!s', ':.4']])

    def __repr__(self):
        """Return ``repr(self)``."""
        posargs = [array_str(self.array)]
        optargs = [('exponent', self.exponent, 2.0)]
        inner_str = signature_string(posargs, optargs, sep=',\n',
                                     mod=['!s', ':.4'])
        return '{}(\n{}\n)'.format(self.__class__.__name__, indent(inner_str))

    def __str__(self):
        """Return ``str(self)``."""
        return repr(self)

    def norm(self, x):
        """Return the weighted norm of ``x``.

        Parameters
        ----------
        x : ArrayLike
            Tensor whose norm is calculated.

        Returns
        -------
        norm : float
            The norm of the provided tensor.
        """
        if self.exponent == 2.0:
            norm_squared = self.inner(x, x).real  # TODO: optimize?!
            if norm_squared < 0:
                norm_squared = 0.0  # Compensate for numerical error
            return float(np.sqrt(norm_squared))
        else:
            return float(_pnorm_diagweight(x, self.exponent, self.array))

    def inner(self, x1, x2):
        """Return the weighted inner product of ``x1`` and ``x2``.

        Parameters
        ----------
        x1, x2 : ArrayLike
            Tensors whose inner product is calculated.

        Returns
        -------
        inner : float or complex
            The inner product of the two provided vectors.
        """
        if self.exponent != 2.0:
            raise NotImplementedError('no inner product defined for '
                                      'exponent != 2 (got {})'
                                      ''.format(self.exponent))
        else:
            inner = _inner_default(x1 * self.array, x2)
            if is_real_dtype(x1.dtype):
                return float(inner)
            else:
                return complex(inner)


class ConstWeighting(Weighting):

    """Weighting of a space by a constant."""

    def __init__(self, const, impl, device, exponent=2.0):
        """Initialize a new instance.

        Parameters
        ----------
        constant : positive float
            Weighting constant of the inner product.
        impl : string
            Specifier for the implementation backend.
        device :
            device identifier, compatible with the backend associated with `impl`
        exponent : positive float, optional
            Exponent of the norm. For values other than 2.0, the inner
            product is not defined.
        """
        super(ConstWeighting, self).__init__(impl=impl, device=device, exponent=exponent)
        self.__const = float(const)

        if self.const <= 0:
            raise ValueError('expected positive constant, got {}'
                             ''.format(const))
        if not np.isfinite(self.const):
            raise ValueError('`const` {} is invalid'.format(const))

    @property
    def const(self):
        """Weighting constant of this inner product."""
        return self.__const

    @property
    def weight(self):
        """Weighting constant of this instance."""
        return self.const

    @property
    def shape(self):
        """A constant weight can be applied to any shape."""
        return ()

    def to_device(self, device):
        return ConstWeighting(const = self.const, impl=self.impl, device=device, exponent=self.exponent)

    def to_impl(self, impl):
        return ConstWeighting(const = self.const, impl=impl, device=self.device, exponent=self.exponent)

    def __eq__(self, other):
        """Return ``self == other``.

        Returns
        -------
        equal : bool
            ``True`` if ``other`` is a `ConstWeighting` instance with the
            same constant, ``False`` otherwise.
        """
        if other is self:
            return True

        return (super(ConstWeighting, self).__eq__(other) and
                self.const == getattr(other, 'const', None))

    def __hash__(self):
        """Return ``hash(self)``."""
        return hash((super(ConstWeighting, self).__hash__(), self.const))

    def equiv(self, other):
        """Test if other is an equivalent weighting.

        Returns
        -------
        equivalent : bool
            ``True`` if other is a `Weighting` instance with the same
            `Weighting.impl`, which yields the same result as this
            weighting for any input, ``False`` otherwise. This is checked
            by entry-wise comparison of matrices/arrays/constants.
        """
        if isinstance(other, ConstWeighting):
            return self == other
        elif isinstance(other, (ArrayWeighting, MatrixWeighting)):
            return other.equiv(self)
        else:
            return False

    @property
    def repr_part(self):
        """String usable in a space's ``__repr__`` method."""
        optargs = [('weighting', self.const, 1.0),
                   ('exponent', self.exponent, 2.0)]
        return signature_string([], optargs, mod=':.4')

    def __repr__(self):
        """Return ``repr(self)``."""
        posargs = [self.const]
        optargs = [('exponent', self.exponent, 2.0)]
        return '{}({})'.format(self.__class__.__name__,
                               signature_string(posargs, optargs))

    def __str__(self):
        """Return ``str(self)``."""
        return repr(self)

    def inner(self, x1, x2):
        """Return the weighted inner product of ``x1`` and ``x2``.

        Parameters
        ----------
        x1, x2 : ArrayLike
            Tensors whose inner product is calculated.

        Returns
        -------
        inner : float or complex
            The inner product of the two provided tensors.
        """
        if self.exponent != 2.0:
            raise NotImplementedError('no inner product defined for '
                                      'exponent != 2 (got {})'
                                      ''.format(self.exponent))
        else:
            return self.const * _inner_default(x1, x2)

    def norm(self, x):
        """Return the weighted norm of ``x``.

        Parameters
        ----------
        x1 : ArrayLike
            Tensor whose norm is calculated.

        Returns
        -------
        norm : float
            The norm of the tensor.
        """
        if self.exponent == 2.0:
            return float(np.sqrt(self.const) * _norm_default(x))
        elif self.exponent == float('inf'):
            return float(self.const * _pnorm_default(x, self.exponent))
        else:
            return float((self.const ** (1 / self.exponent) *
                          _pnorm_default(x, self.exponent)))

    def dist(self, x1, x2):
        """Return the weighted distance between ``x1`` and ``x2``.

        Parameters
        ----------
        x1, x2 : `NumpyTensor`
            Tensors whose mutual distance is calculated.

        Returns
        -------
        dist : float
            The distance between the tensors.
        """
        if self.exponent == 2.0:
            return float(np.sqrt(self.const) * _norm_default(x1 - x2))
        elif self.exponent == float('inf'):
            return float(self.const * _pnorm_default(x1 - x2, self.exponent))
        else:
            return float((self.const ** (1 / self.exponent) *
                          _pnorm_default(x1 - x2, self.exponent)))


class CustomInner(Weighting):

    """Class for handling a user-specified inner product."""

    def __init__(self, inner, impl, device, shape=()):
        """Initialize a new instance.

        Parameters
        ----------
        inner : callable
            The inner product implementation. It must accept two
            `LinearSpaceElement` arguments, return an element from
            their space's field (real or complex number) and
            satisfy the following conditions for all space elements
            ``x, y, z`` and scalars ``s``:

            - ``<x, y> = conj(<y, x>)``
            - ``<s*x + y, z> = s * <x, z> + <y, z>``
            - ``<x, x> = 0``  if and only if  ``x = 0``

        impl : string
            Specifier for the implementation backend.
        device :
            device identifier, compatible with the backend associated with `impl`
        shape :
            what shape array need to have to be processed by this weighting.
            The `inner` callable can assume that the shape has already been checked.
            If an empty shape is specified (the default), `inner` should be able to
            handle arrays of arbitrary shape.
        """
        super(CustomInner, self).__init__(impl=impl, device=device, exponent=2.0)
        
        self.__shape = shape

        if not callable(inner):
            raise TypeError('`inner` {!r} is not callable'
                            ''.format(inner))
        self.__inner = inner

    @property
    def shape(self):
        return self.__shape

    @property
    def inner(self):
        """Custom inner product of this instance.."""
        return self.__inner

    def __eq__(self, other):
        """Return ``self == other``.

        Returns
        -------
        equal : bool
            ``True`` if other is a `CustomInner`
            instance with the same inner product, ``False`` otherwise.
        """
        return (super(CustomInner, self).__eq__(other) and
                self.inner == other.inner)

    def __hash__(self):
        """Return ``hash(self)``."""
        return hash((super(CustomInner, self).__hash__(), self.inner))

    @property
    def repr_part(self):
        """String usable in a space's ``__repr__`` method."""
        optargs = [('inner', self.inner, '')]
        return signature_string([], optargs, mod='!r')

    def __repr__(self):
        """Return ``repr(self)``."""
        posargs = [self.inner]
        optargs = []
        inner_str = signature_string(posargs, optargs, mod='!r')
        return '{}({})'.format(self.__class__.__name__, inner_str)


class CustomNorm(Weighting):

    """Class for handling a user-specified norm.

    Note that this removes ``inner``.
    """

    def __init__(self, norm, impl, device, shape=()):
        """Initialize a new instance.

        Parameters
        ----------
        norm : callable
            The norm implementation. It must accept a
            `LinearSpaceElement` argument, return a float and satisfy
            the following conditions for all space elements
            ``x, y`` and scalars ``s``:

            - ``||x|| >= 0``
            - ``||x|| = 0``  if and only if  ``x = 0``
            - ``||s * x|| = |s| * ||x||``
            - ``||x + y|| <= ||x|| + ||y||``
        impl : string
            Specifier for the implementation backend
        device :
            device identifier, compatible with the backend associated with `impl`
        shape :
            what shape array need to have to be processed by this weighting.
            The `norm` callable can assume that the shape has already been checked.
            If an empty shape is specified (the default), `norm` should be able to
            handle arrays of arbitrary shape.
        """
        super(CustomNorm, self).__init__(impl=impl, device=device, exponent=1.0)

        self.__shape = shape

        if not callable(norm):
            raise TypeError('`norm` {!r} is not callable'
                            ''.format(norm))
        self.__norm = norm

    @property
    def shape(self):
        return self.__shape

    def inner(self, x1, x2):
        """Inner product is not defined for custom distance."""
        raise NotImplementedError('`inner` not defined for custom norm')

    @property
    def norm(self):
        """Custom norm of this instance.."""
        return self.__norm

    def __eq__(self, other):
        """Return ``self == other``.

        Returns
        -------
        equal : bool
            ``True`` if other is a `CustomNorm` instance with the same
            norm, ``False`` otherwise.
        """
        return (super(CustomNorm, self).__eq__(other) and
                self.norm == other.norm)

    def __hash__(self):
        """Return ``hash(self)``."""
        return hash((super(CustomNorm, self).__hash__(), self.norm))

    @property
    def repr_part(self):
        """Return a string usable in a space's ``__repr__`` method."""
        optargs = [('norm', self.norm, ''),
                   ('exponent', self.exponent, 2.0)]
        return signature_string([], optargs, mod=[[], ['!r', ':.4']])

    def __repr__(self):
        """Return ``repr(self)``."""
        posargs = [self.norm]
        optargs = [('exponent', self.exponent, 2.0)]
        inner_str = signature_string(posargs, optargs, mod=['!r', ':.4'])
        return '{}({})'.format(self.__class__.__name__, inner_str)


class CustomDist(Weighting):

    """Class for handling a user-specified distance.

    Note that this removes ``inner`` and ``norm``.
    """

    def __init__(self, dist, impl, device, shape=()):
        """Initialize a new instance.

        Parameters
        ----------
        dist : callable
            The distance function defining a metric on a `LinearSpace`.
            It must accept two `LinearSpaceElement` arguments, return a
            float and and fulfill the following mathematical conditions
            for any three space elements ``x, y, z``:

            - ``dist(x, y) >= 0``
            - ``dist(x, y) = 0``  if and only if  ``x = y``
            - ``dist(x, y) = dist(y, x)``
            - ``dist(x, y) <= dist(x, z) + dist(z, y)``
        impl : string
            Specifier for the implementation backend
        device :
            device identifier, compatible with the backend associated with `impl`
        shape :
            what shape array need to have to be processed by this weighting.
            The `dist` callable can assume that the shape has already been checked.
            If an empty shape is specified (the default), `dist` should be able to
            handle arrays of arbitrary shape.
        """
        super(CustomDist, self).__init__(impl=impl, device=device, exponent=1.0)

        self.__shape = shape

        if not callable(dist):
            raise TypeError('`dist` {!r} is not callable'
                            ''.format(dist))
        self.__dist = dist

    @property
    def shape(self):
        return self.__shape

    @property
    def dist(self):
        """Custom distance of this instance.."""
        return self.__dist

    def inner(self, x1, x2):
        """Inner product is not defined for custom distance."""
        raise NotImplementedError('`inner` not defined for custom distance')

    def norm(self, x):
        """Norm is not defined for custom distance."""
        raise NotImplementedError('`norm` not defined for custom distance')

    def __eq__(self, other):
        """Return ``self == other``.

        Returns
        -------
        equal : bool
            ``True`` if other is a `CustomDist` instance with the same
            dist, ``False`` otherwise.
        """
        return (super(CustomDist, self).__eq__(other) and
                self.dist == other.dist)

    def __hash__(self):
        """Return ``hash(self)``."""
        return hash((super(CustomDist, self).__hash__(), self.dist))

    @property
    def repr_part(self):
        """Return a string usable in a space's ``__repr__`` method."""
        optargs = [('dist', self.dist, '')]
        return signature_string([], optargs, mod=['', '!r'])

    def __repr__(self):
        """Return ``repr(self)``."""
        posargs = [self.dist]
        optargs = []
        inner_str = signature_string(posargs, optargs, mod=['!r', ''])
        return '{}({})'.format(self.__class__.__name__, inner_str)


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()
